import pytest

from synthegrator.environments import ProjectWorkingDirectory
from synthegrator.memory_fs import ProjectDir, ProjectFile
from synthegrator.transformation_spec import (
    MarkElement,
    MarkText,
    StsEditable,
    StsInsert,
    StsInvisible,
    StsNormalText,
    StsPath,
    StsPrefixCommentSelector,
    StsSelectorCharSpan,
    StsSelectorLineSpan,
    StsSpecStatement,
    StsValueQuery,
    TransformationSpec,
    markup_path,
)


def test_glob_select1():
    selector = StsPath(path_glob="foo.py")
    assert selector.does_match_path("foo.py")


def test_glob_select2():
    selector = StsPath(path_glob="*.py")
    assert selector.does_match_path("foo.py")
    assert not selector.does_match_path("foo.java")


def test_glob_select3():
    selector = StsPath(path_glob="foo/*.py")
    assert selector.does_match_path("foo/bar.py")
    assert not selector.does_match_path("foo.py")


def test_split_text():
    text = "bar"
    node = MarkText(text, 0, len(text))
    assert node.text == text
    assert node.mark_start_original_char == 0
    assert node.mark_end_original_char == 3
    first, second = node.split(1)
    assert first == MarkText("b", 0, 1)
    assert second == MarkText("ar", 1, 3)
    assert second.mark_start_original_char == 1
    assert second.mark_end_original_char == 3


def test_split_text1():
    text = "bar"
    start_offset = 5
    node = MarkText(text, start_offset, start_offset + len(text))
    first, second = node.split(start_offset + 2)
    assert first == MarkText("ba", start_offset, start_offset + 2)
    assert second == MarkText("r", start_offset + 2, start_offset + 3)


def test_split_text2():
    text = "bar"
    start_offset = 0
    node = MarkText(text, start_offset, start_offset + len(text))
    first, second = node.split(0)
    assert first is None
    assert second == node


def test_split_text3():
    text = "bar"
    start_offset = 0
    node = MarkText(text, start_offset, start_offset + len(text))
    first, second = node.split(3)
    assert first == node
    assert second is None


def test_split_span():
    text = "bar"
    start_offset = 0
    node = MarkText(text, start_offset, start_offset + len(text))
    before, inner, after = node.split_span(1, 2)
    assert before == MarkText("b", start_offset, start_offset + 1)
    assert inner == MarkText("a", start_offset + 1, start_offset + 2)
    assert after == MarkText("r", start_offset + 2, start_offset + 3)


def test_split_span_before():
    text = "bar"
    start_offset = 5
    node = MarkText(text, start_offset, start_offset + len(text))
    before, inner, after = node.split_span(0, start_offset + 1)
    assert before is None
    assert inner == MarkText("b", start_offset, start_offset + 1)
    assert after == MarkText("ar", start_offset + 1, start_offset + 3)


def test_split_span_all():
    text = "bar"
    start_offset = 5
    node = MarkText(text, start_offset, start_offset + len(text))
    before, inner, after = node.split_span(0, 99)
    assert before is None
    assert inner == MarkText("bar", start_offset, start_offset + 3)
    assert after is None


def test_insert_node1():
    text = "bar"
    node = MarkElement(StsNormalText(), [MarkText(text, 0, 3)], 0, 3)
    node.insert_node(StsInvisible(), 1, 2)
    print("\n")
    print(node.as_xml())
    assert node.children[0] == MarkText("b", 0, 1)
    assert node.children[1] == MarkElement(
        StsInvisible(),
        [MarkText(text[1:2], 1, 2)],
        1,
        2,
    )
    assert node.children[2] == MarkText("r", 2, 3)


def test_insert_node_left_side():
    text = "bar"
    node = MarkElement(StsNormalText(), [MarkText(text, 0, 3)], 0, 3)
    node.insert_node(StsInvisible(), 0, 2)
    print(node.as_xml())
    assert len(node.children) == 2
    assert node.children[0] == MarkElement(
        StsInvisible(),
        [MarkText(text[0:2], 0, 2)],
        0,
        2,
    )
    assert node.children[1] == MarkText("r", 2, 3)


def test_insert_node_right_side():
    text = "bar"
    node = MarkElement(StsNormalText(), [MarkText(text, 0, 3)], 0, 3)
    node.insert_node(StsInvisible(), 1, 3)
    print(node.as_xml())
    assert len(node.children) == 2
    assert node.children[0] == MarkText("b", 0, 1)
    assert node.children[1] == MarkElement(
        StsInvisible(),
        [MarkText(text[1:3], 1, 3)],
        1,
        3,
    )


def test_insert_node_full():
    text = "bar"
    node = MarkElement(StsNormalText(), [MarkText(text, 0, 3)], 0, 3)
    node.insert_node(StsInvisible(), 0, 3)
    print(node.as_xml())
    assert len(node.children) == 1
    assert node.children[0] == MarkElement(StsInvisible(), [MarkText(text, 0, 3)], 0, 3)


def test_too_far_past():
    text = "bar"
    node = MarkElement(StsNormalText(), [MarkText(text, 0, 3)], 0, 3)
    with pytest.raises(ValueError):
        node.insert_node(StsInvisible(), 0, 5)


def test_too_single():
    text = "bar"
    node = MarkElement(StsNormalText(), [MarkText(text, 0, 3)], 0, 3)
    node.insert_node(StsInvisible(), 1, 1)
    print("\n", node.as_xml())
    assert len(node.children) == 3
    assert node.children[0] == MarkText("b", 0, 1)
    assert node.children[1] == MarkElement(StsInvisible(), [], 1, 1)
    assert node.children[2] == MarkText("ar", 1, 3)


def test_too_single_end():
    text = "bar"
    node = MarkElement(StsNormalText(), [MarkText(text, 0, 3)], 0, 3)
    node.insert_node(StsInvisible(), 3, 3)
    print("\n", node.as_xml())
    assert len(node.children) == 2
    assert node.children[0] == MarkText("bar", 0, 3)
    assert node.children[1] == MarkElement(StsInvisible(), [], 3, 3)


def test_multi_element():
    node = MarkElement(
        StsNormalText(),
        [
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkText("baz", 6, 9),
        ],
        0,
        9,
    )
    node.insert_node(StsEditable(), 3, 6)
    print("\n", node.as_xml())
    assert len(node.children) == 3
    assert node.children[0] == MarkText("foo", 0, 3)
    val = node.children[1]
    expected = MarkElement(
        StsInvisible(),
        [MarkElement(StsEditable(), [MarkText("bar", 3, 6)], 3, 6)],
        3,
        6,
    )
    assert val.as_xml() == expected.as_xml()
    assert val == expected
    assert node.children[2] == MarkText("baz", 6, 9)


def test_multi_element_cut():
    node = MarkElement(
        StsNormalText(),
        [
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkText("baz", 6, 9),
        ],
        0,
        9,
    )
    node.insert_node(StsEditable(), 2, 6)
    print("\n", node.as_xml())
    assert len(node.children) == 3
    assert node.children[0] == MarkText("fo", 0, 2)
    val = node.children[1]
    expected = MarkElement(
        StsEditable(),
        [
            MarkText("o", 2, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
        ],
        2,
        6,
    )
    assert val.as_xml() == expected.as_xml()
    assert val == expected
    assert node.children[2] == MarkText("baz", 6, 9)


def test_multi_element_cut_twice():
    node = MarkElement(
        StsNormalText(),
        [
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkText("baz", 6, 9),
        ],
        0,
        9,
    )
    node.insert_node(StsEditable(), 2, 7)
    print("\n", node.as_xml())
    assert len(node.children) == 3
    assert node.children[0] == MarkText("fo", 0, 2)
    val = node.children[1]
    expected = MarkElement(
        StsEditable(),
        [
            MarkText("o", 2, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkText("b", 6, 7),
        ],
        2,
        7,
    )
    assert val.as_xml() == expected.as_xml()
    assert val == expected
    assert node.children[2] == MarkText("az", 7, 9)


def test_multi_element_bad():
    node = MarkElement(
        StsNormalText(),
        [
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkText("baz", 6, 9),
        ],
        0,
        9,
    )
    with pytest.raises(ValueError):
        node.insert_node(StsEditable(), 2, 5)
        print("\n", node.as_xml())


def test_multi_element_many_elements():
    node = MarkElement(
        StsNormalText(),
        [
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkElement(StsInvisible(), [MarkText("pop", 6, 9)], 6, 9),
            MarkElement(StsInvisible(), [MarkText("bap", 9, 12)], 9, 12),
            MarkText("baz", 12, 15),
        ],
        0,
        15,
    )
    node.insert_node(StsEditable(), 3, 9)
    assert len(node.children) == 4
    assert node.children[0] == MarkText("foo", 0, 3)
    val = node.children[1]
    expected = MarkElement(
        StsEditable(),
        [
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkElement(StsInvisible(), [MarkText("pop", 6, 9)], 6, 9),
        ],
        3,
        9,
    )
    assert val.as_xml() == expected.as_xml()
    assert val == expected
    assert node.children[2] == MarkElement(
        StsInvisible(),
        [MarkText("bap", 9, 12)],
        9,
        12,
    )
    assert node.children[3] == MarkText("baz", 12, 15)


def test_markup_path():
    wd = ProjectWorkingDirectory(
        files=ProjectDir(
            {
                "foo.txt": ProjectFile("foo.txt", "0123456789"),
            },
        ),
    )
    spec = TransformationSpec(
        (
            StsSpecStatement(
                StsPath("*.txt"),
                StsSelectorCharSpan(1, 3),
                StsInvisible(),
            ),
            StsSpecStatement(
                StsPath("foo.py"),
                StsSelectorCharSpan(0, 9),
                StsInvisible(),
            ),  # Not match
            StsSpecStatement(StsPath("*.txt"), StsSelectorCharSpan(5, 5), StsInsert()),
        ),
    )
    node = markup_path(wd, "foo.txt", spec)
    node.pretty_print_xml()
    expected = MarkElement(
        StsNormalText(),
        [
            MarkText("0", 0, 1),
            MarkElement(StsInvisible(), [MarkText("12", 1, 3)], 1, 3),
            MarkText("34", 3, 5),
            MarkElement(StsInsert(), [], 5, 5),
            MarkText("56789", 5, 10),
        ],
        0,
        10,
    ).freeze()
    assert node.as_xml() == expected.as_xml()
    assert repr(node) == repr(expected)
    assert node == expected


def test_markup_path_double_sided_part():
    wd = ProjectWorkingDirectory(
        files=ProjectDir(
            {
                "foo.txt": ProjectFile("foo.txt", "0\n1\n2\n3\n4"),
            },
        ),
    )
    spec = TransformationSpec(
        (
            StsSpecStatement(
                StsPath("*.txt"),
                StsSelectorLineSpan(0, 1),
                StsInvisible(),
            ),
            StsSpecStatement(
                StsPath("*.txt"),
                StsSelectorLineSpan(4, 5),
                StsInvisible(),
            ),
            # StsSpecStatement(
            # ),
        ),
    )
    node = markup_path(wd, "foo.txt", spec)
    node.pretty_print_xml()
    expected = MarkElement(
        StsNormalText(),
        [
            MarkElement(StsInvisible(), [MarkText("0\n", 0, 2)], 0, 2),
            MarkText("1\n2\n3\n", 2, 8),
            MarkElement(StsInvisible(), [MarkText("4", 8, 9)], 8, 9),
        ],
        0,
        9,
    ).freeze()
    assert node.as_xml() == expected.as_xml()
    assert repr(node) == repr(expected)
    assert node == expected


def test_markup_path_double_sided_full():
    wd = ProjectWorkingDirectory(
        files=ProjectDir(
            {
                "foo.txt": ProjectFile("foo.txt", "0\n1\n2\n3\n4"),
            },
        ),
    )
    spec = TransformationSpec(
        (
            StsSpecStatement(
                StsPath("*.txt"),
                StsSelectorLineSpan(0, 1),
                StsInvisible(),
            ),
            StsSpecStatement(
                StsPath("*.txt"),
                StsSelectorLineSpan(4, 5),
                StsInvisible(),
            ),
            StsSpecStatement(
                StsPath("*.txt"),
                StsSelectorLineSpan(2, 3),
                StsEditable(),
            ),
        ),
    )
    node = markup_path(wd, "foo.txt", spec)
    node.pretty_print_xml()
    expected = MarkElement(
        StsNormalText(),
        [
            MarkElement(StsInvisible(), [MarkText("0\n", 0, 2)], 0, 2),
            MarkText("1\n", 2, 4),
            MarkElement(StsEditable(), [MarkText("2\n", 4, 6)], 4, 6),
            MarkText("3\n", 6, 8),
            MarkElement(StsInvisible(), [MarkText("4", 8, 9)], 8, 9),
        ],
        0,
        9,
    ).freeze()
    assert node.as_xml() == expected.as_xml()
    assert repr(node) == repr(expected)
    assert node == expected


def test_line_selector():
    selector = StsSelectorLineSpan(1, 3)
    text = "foo\nbar\nbaz\npop"
    spans = selector.get_matching_spans(text, path="foo.txt")
    assert len(spans) == 1
    start, end = spans[0]
    assert text[start:end] == "bar\nbaz\n"


def test_line_selector2():
    selector = StsSelectorLineSpan(0, 2)
    text = "foob\nbarss\nbaz\npop"
    spans = selector.get_matching_spans(text, path="foo.txt")
    assert len(spans) == 1
    start, end = spans[0]
    assert text[start:end] == "foob\nbarss\n"


def test_line_selector3():
    selector = StsSelectorLineSpan(3, 4)
    text = "foob\nbarss\nbaz\npop"
    spans = selector.get_matching_spans(text, path="foo.txt")
    assert len(spans) == 1
    start, end = spans[0]
    assert text[start:end] == "pop"


def test_static_verb():
    verb = StsValueQuery()
    assert verb.operation_is_static()
    statement = StsSpecStatement(
        StsPath("*.py"),
        StsPrefixCommentSelector(),
        verb=verb,
    )
    assert statement.operation_is_static()
    spec = TransformationSpec((statement,))
    file = (
        "def foo():\n"
        '    """my comment"""\n'
        "    pass\n"
        "\n"
        "def bar():\n"
        '    """my comment"""\n'
        "    pass\n"
    )
    wd = ProjectWorkingDirectory(
        files=ProjectDir.construct_with_files([("foo.py", file)]),
    )
    node = markup_path(wd, "foo.py", spec)
    print(node.as_xml())
    # This is copied from whatever was printed. Things
    #   like the random hash might reasonably change.
    #   The main thing to confirm is that the two ids of the
    #   Value Queries are the same.
    assert (
        node.as_xml()
        == '''<StsNormalText id="FsD6y" _original_start_char="0" _original_end_char="83">def foo():
<StsValueQuery id="s_3Ypr9A" _original_start_char="11" _original_end_char="32">    """my comment"""
</StsValueQuery>    pass

def bar():
<StsValueQuery id="s_3Ypr9A" _original_start_char="53" _original_end_char="74">    """my comment"""
</StsValueQuery>    pass
</StsNormalText>'''
    )



from collections.abc import Iterator
from enum import Enum
from functools import partial
from itertools import islice

from tqdm import tqdm

from synthegrator.code_problem_builders import (
    make_instruction_only_problem,
    make_simple_method_completion_problem,
)
from synthegrator.code_problems import (
    CodeProblem,
    TestCase,
    TestCaseArbitraryMethod,
)
from synthegrator.lang_specs.lang_spec import LsFunctionDef
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.problem_rendering import render_single_edit_codeproblem_autoregressive
from synthegrator.response_parser import set_base_indent
from synthegrator.solution_eval import evaluate_code_problem_execution
from synthegrator.synthdatasets import load_hf_dataset_with_backoff
from synthegrator.transformation_spec import (
    StsInsert,
)


class MbppSplit(Enum):
    TRAIN = "train"
    VALIDATION = "validation"
    TEST = "test"
    PROMPT = "prompt"
    ALL = "all"
    ALL_BUT_PROMPT = "all_but_prompt"
    """The mbpp dataset includes a split 10 problems that was their prompt for
    few shot solving. One might want to not do few shotting (or with alterative 
    leave-few-out versions), so we default to ALL including these. Use this version
    to exclude the 10 problems.
    """
    EVAL_PLUS = "eval_plus"
    """Load evalplus dataset with more test cases (evalplus.github.io). 
    Note though, this is only 374 problems, rather than the 974 
    found in actual MBPP"""


def yield_mbpp(
    split: MbppSplit = MbppSplit.ALL,
    reformat_whitespace: bool = True,
    sanitized_only: bool = False,
    only_with_passing_reference_solution: bool = False,
    format_as_method_completion_problem: bool = False,
    max_problems: int = None,
) -> Iterator[CodeProblem]:
    """
    :param split: which split to yield
    :param reformat_whitespace: whether to convert tabs to spaces and convert to unix line endings
    :param sanitized_only: whether to only yield sanitized examples. These are
        hand-verified by the authors.
    :param only_with_passing_reference_solution: About ~80/~970 of the problems don't
        seem to have a reference solution that does not pass the provided tests.
        You can set this to true to filter.
    :param format_as_method_completion_problem: Normally mbpp is file generation
        from instructions. Occasionally (about ~60 problems) this involves
        implemented a class or multiple methods. Setting this parameter will
        instead make the problem an insertion problem to fill in the contents
        of main method under test. Other context (like other classes, methods
        or imports) are provided as visible to the solver. This can make it
        easier to the solver as things like the imports can be a hint. It also
        makes it easier to get autoregressive language models to cooperate, as the
        method completion will always terminate at a block-end, whereas when
        just generating arbitrary completions from a model can end up with
        a lot of extraneous output that can be more difficult to filter out.
    :param max_problems: maximum number of problems to yield
    """
    # The all and the santized use different schemas
    dataset_all = _get_mbpp_huggingface(sanitized=False)
    """
    DatasetDict({
        test: Dataset({
            features: ['task_id', 'text', 'code', 'test_list', 'test_setup_code', 'challenge_test_list'],
            num_rows: 974
        })
    })
    """
    dataset_sanitized = _get_mbpp_huggingface(sanitized=True)
    """
    DatasetDict({
        test: Dataset({
            features: ['source_file', 'task_id', 'prompt', 'code', 'test_imports', 'test_list'],
            num_rows: 427
        })
    })
    """
    sanitized_key_to_instance = {
        ex["task_id"]: ex for ex in _iterate_for_split(dataset_sanitized, split)
    }

    exclude_task_ids = (
        _task_ids_where_reference_solution_fails()
        if only_with_passing_reference_solution
        else set()
    )

    produce_func = (
        _produce_code_problem_for_mbpp_method_completion
        if format_as_method_completion_problem
        else _produce_code_problem_for_mbpp
    )

    yielded_prob_count = 0
    for ex in _iterate_for_split(dataset_all, split):
        if max_problems is not None and yielded_prob_count >= max_problems:
            break
        santized_version = sanitized_key_to_instance.get(ex["task_id"], {})
        if sanitized_only and len(santized_version) == 0:
            continue
        if str(ex["task_id"]) in exclude_task_ids:
            continue
        try:
            prob = produce_func(
                prompt=santized_version.get("prompt", ex["text"]),
                code=ex["code"],
                challenge_test_list=ex["challenge_test_list"],
                task_id=ex["task_id"],
                test_list=santized_version.get("test_list", ex["test_list"]),
                test_setup_code=ex["test_setup_code"],
                test_imports=santized_version.get("test_imports", None),
                reformat_whitespace=reformat_whitespace,
                is_in_santized=len(santized_version) > 0,
            )
            yield prob
            yielded_prob_count += 1
        except DatasetParseException:
            # raise
            continue


def yield_mbpp_plus(
    reformat_whitespace: bool = True,
    format_as_method_completion_problem: bool = False,
    max_problems: int = None,
):
    """
    Load the MBPP+ dataset (evalplus.github.io). This dataset is intended
    to have more rigorous test cases than the original MBPP dataset. However,
    it has the disadvantage of being smaller (374 problems) compared to the
    974 problems in all splits of the original MBPP dataset.
    """
    ds = _get_mbpp_plus_huggingface()
    produce_func = (
        _produce_code_problem_for_mbpp_method_completion
        if format_as_method_completion_problem
        else _produce_code_problem_for_mbpp
    )
    assert list(ds.keys()) == ["test"]

    num_yielded = 0
    for ex in ds["test"]:
        if max_problems is not None and num_yielded >= max_problems:
            break
        try:
            prob = produce_func(
                prompt=ex["prompt"],
                code=ex["code"],
                challenge_test_list=[],
                task_id=ex["task_id"],
                test_preformated_src=ex["test"],
                test_list=ex["test_list"],
                test_setup_code="",
                test_imports=ex.get("test_imports", None),
                reformat_whitespace=reformat_whitespace,
                is_in_santized=False,
                is_plus=True,
            )
            assert prob.dataset_name == "mbpp_plus"
            yield prob
            num_yielded += 1
        except DatasetParseException as e:
            # raise
            continue


def _get_mbpp_huggingface(
    sanitized: bool,
):
    if not sanitized:
        return load_hf_dataset_with_backoff("mbpp", "full")
    return load_hf_dataset_with_backoff(
        "mbpp",
        "sanitized",
    )


def _get_mbpp_plus_huggingface():
    data = load_hf_dataset_with_backoff("evalplus/mbppplus")
    return data


def _iterate_for_split(dataset, split: MbppSplit):
    all_main = split in {MbppSplit.ALL, MbppSplit.ALL_BUT_PROMPT}
    if split == MbppSplit.TRAIN or all_main:
        yield from dataset["train"]
    if split == MbppSplit.VALIDATION or all_main:
        yield from dataset["validation"]
    if split == MbppSplit.TEST or all_main:
        yield from dataset["test"]
    if split in {MbppSplit.PROMPT, MbppSplit.ALL}:
        yield from dataset["prompt"]


def _task_ids_where_reference_solution_fails():
    return {
        "602",
        "617",
        "620",
        "626",
        "627",
        "631",
        "633",
        "643",
        "721",
        "734",
        "737",
        "743",
        "746",
        "747",
        "751",
        "755",
        "756",
        "757",
        "760",
        "763",
        "767",
        "770",
        "777",
        "781",
        "782",
        "787",
        "793",
        "794",
        "797",
        "799",
        "801",
        "802",
        "803",
        "804",
        "521",
        "576",
        "586",
        "597",
        "16",
        "19",
        "46",
        "51",
        "57",
        "61",
        "70",
        "75",
        "92",
        "101",
        "117",
        "119",
        "125",
        "130",
        "137",
        "145",
        "159",
        "160",
        "164",
        "166",
        "167",
        "172",
        "245",
        "264",
        "270",
        "285",
        "296",
        "327",
        "406",
        "433",
        "434",
        "436",
        "438",
        "452",
        "454",
        "468",
        "471",
    }


_mbpp_insert_node = StsInsert(stop_at_block_end=False)


def _produce_code_problem_for_mbpp(
    prompt: str,
    code: str,  # solution code
    challenge_test_list: list[str],
    task_id: int,
    test_list: list[str],
    test_setup_code: str,
    reformat_whitespace: bool,
    test_imports: list[str],
    is_in_santized: bool,
    test_preformated_src: str = None,
    is_plus: bool = False,
) -> CodeProblem:
    # Copy prompt from https://github.com/google-research/google-research/tree/master/mbpp
    # You are an expert Python programmer, and here is your task: {prompt} Your code should pass these tests:\n\n{tests}\n[BEGIN]\n{code}\n[DONE
    test_prompt = "\n".join(test_list)
    prompt = (
        f"You are an expert Python programmer, and here is your task: {prompt} "
        "Your code should pass these tests:\n\n"
        f"{test_prompt}\n\n"
    )
    if reformat_whitespace:
        prompt = _reformat_whitespace(prompt)

    if not test_preformated_src:
        test_cases = _make_mbpp_test_cases(
            test_list,
            challenge_test_list,
            test_imports,
            test_setup_code,
            reformat_whitespace,
        )
    else:
        test_cases = _make_mbpp_plus_test_cases(
            test_list,
            challenge_test_list,
            test_imports,
            test_setup_code,
            reformat_whitespace,
            test_reformated_src=test_preformated_src,
        )

    return make_instruction_only_problem(
        instructions=prompt,
        insert_node=_mbpp_insert_node,
        test_cases=test_cases,
        dataset_name="mbpp" if not is_plus else "mbpp_plus",
        problem_id=str(task_id),
        known_solutions=[code],
    )


def _produce_code_problem_for_mbpp_method_completion(
    prompt: str,
    code: str,  # solution code
    challenge_test_list: list[str],
    task_id: int,
    test_list: list[str],
    test_setup_code: str,
    reformat_whitespace: bool,
    test_imports: list[str],
    is_in_santized: bool,
    test_preformated_src: str = None,
    is_plus: bool = False,
) -> CodeProblem:
    test_prompt = "\n".join(test_list)
    reformat = _reformat_whitespace if reformat_whitespace else lambda x: x
    prompt = reformat(f"{prompt}\n\nExamples:\n{test_prompt}")
    ref_func = _get_reference_method_under_test(reformat(code), test_list)
    if not test_preformated_src:
        test_cases = _make_mbpp_test_cases(
            test_list,
            challenge_test_list,
            test_imports,
            test_setup_code,
            reformat_whitespace,
        )
    else:
        test_cases = _make_mbpp_plus_test_cases(
            test_list,
            challenge_test_list,
            test_imports,
            test_setup_code,
            reformat_whitespace,
            test_reformated_src=test_preformated_src,
        )

    return make_simple_method_completion_problem(
        prompt=prompt,
        signature=reformat(ref_func.get_declaration_str()),
        model_context=reformat(ref_func.get_preceding_src_code()).lstrip("\n"),
        known_solutions=[
            set_base_indent(reformat(ref_func.get_body_src()).lstrip("\n"), 4),
        ],
        after_method_src=ref_func.get_src_code_after().strip(),
        test_cases=test_cases,
        dataset_name="mbpp" if not is_plus else "mbpp_plus",
        problem_id=str(task_id) + "func",
    )


class DatasetParseException(Exception):
    pass


def _get_reference_method_under_test(reference_code, test_list) -> LsFunctionDef:
    method_name = _get_mbpp_method_name_under_test(test_list)
    if not method_name:
        msg = (
            f"Expected to find method name.\nReference code:\n{reference_code}\nTest"
            f" list:\n{test_list}"
        )
        raise DatasetParseException(
            msg,
        )
    funcs = list(PythonLangSpec().find_functions(reference_code))
    for func in funcs:
        if func.get_function_name() == method_name:
            return func
    msg = f"Unable to find the mbpp method under test from the test list:\n{test_list}"
    raise DatasetParseException(
        msg,
    )


def _get_mbpp_method_name_under_test(test_list: list[str]) -> str:
    def _guess_from_assert_statement(statement: str):
        if statement.startswith("assert ("):
            # A weird case where just has an unnessary paren (ex, 769)
            statement = statement[len("assert (") :]
        else:
            statement = statement[len("assert ") :]
        # just guess that it is before the first open paren. This could
        #   fail if wrapped in some other call, but will see if the same for all the tests
        return statement.split("(", 1)[0].strip()

    guesses = {_guess_from_assert_statement(stmt) for stmt in test_list}
    if len(guesses) != 1:
        msg = (
            "Unable to guess the mbpp method under test from the test"
            f" list:\n{test_list}"
        )
        raise DatasetParseException(
            msg,
        )
    value = guesses.pop()
    common_builtins = {
        "max",
        "min",
        "list",
        "tuple",
        "str",
        "dict",
        "set",
        "len",
        "sum",
        "sorted",
        "reversed",
        "enumerate",
    }
    if value in common_builtins:
        msg = (
            "The mbpp method under test is a common builtin. This is probably a bad"
            " parse."
        )
        raise DatasetParseException(
            msg,
        )
    return value


def _make_mbpp_test_cases(
    test_list,
    challenge_test_list,
    test_imports,
    test_setup_code,
    reformat_whitespace,
):
    def make_tc(tc_list, is_hidden):
        return _parse_test_cases(
            test_cases=tc_list,
            test_case_imports=test_imports,
            test_case_setup=test_setup_code,
            is_hidden_test=is_hidden,
            reformat_whitespace=reformat_whitespace,
        )

    return make_tc(test_list, False) + make_tc(challenge_test_list, True)


def _make_mbpp_plus_test_cases(
    test_list,
    challenge_test_list,
    test_imports,
    test_setup_code,
    reformat_whitespace,
    test_reformated_src,
):
    base_tests = _make_mbpp_test_cases(
        test_list, [], test_imports, test_setup_code, reformat_whitespace
    )
    challenge = [
        TestCaseArbitraryMethod(
            is_hidden_test=True,
            fail_message="Fail challenge mbpp test",
            method_body=test_reformated_src,
            lang_spec=PythonLangSpec(),
        )
    ]
    return base_tests + challenge


def _reformat_whitespace(code: str) -> str:
    """Converts tabs to spaces and changes to unix line endings"""
    return code.replace("\t", "    ").replace("\r\n", "\n")


def _parse_test_cases(
    test_cases: list[str],
    test_case_imports: list[str],
    test_case_setup: str,
    is_hidden_test: bool,
    reformat_whitespace: bool,
) -> list[TestCase]:
    out = []
    for tc in test_cases:
        tc_imports = "\n".join(test_case_imports or [])
        body = f"{tc_imports}\n{test_case_setup}\n{tc}"
        if reformat_whitespace:
            body = _reformat_whitespace(body)
        body = body.lstrip()
        out.append(
            TestCaseArbitraryMethod(
                is_hidden_test=is_hidden_test,
                fail_message=(
                    f"Fail {'hidden' if is_hidden_test else 'unhidden'} test case: {tc}"
                ),
                method_body=body,
                lang_spec=PythonLangSpec(),
            ),
        )
    return out


def print_problems_with_failing_reference_solutions():
    bad_probs = []
    for problem in tqdm(
        list(
            yield_mbpp(
                split=MbppSplit.ALL,
                sanitized_only=False,
                only_with_passing_reference_solution=False,
            ),
        ),
    ):
        if len(problem.known_solutions) == 0:
            bad_probs.append(problem)
        result = evaluate_code_problem_execution(
            problem,
            problem.known_solutions[0],
            do_not_execute_if_syntax_fail=True,
        )
        if not result.all_passed():
            bad_probs.append(problem)
    print("Problems with failing reference solutions:")
    prob_ids = [p.problem_id for p in bad_probs]
    print(repr(prob_ids))
    assert set(prob_ids) == set(_task_ids_where_reference_solution_fails())
    print("Matches current list")


def _eval_failing():
    fails = [
        prob
        for prob in yield_mbpp(
            split=MbppSplit.ALL,
            sanitized_only=False,
            only_with_passing_reference_solution=False,
        )
        if prob.problem_id in _task_ids_where_reference_solution_fails()
    ]
    for problem in tqdm(fails):
        r = evaluate_code_problem_execution(
            problem,
            problem.known_solutions[0],
            do_not_execute_if_syntax_fail=True,
        )
        print("Instructions:")
        print(problem.instructions)
        print("Render:")
        print(render_single_edit_codeproblem_autoregressive(problem).text)
        assert not r.all_passed()


def _measure_method_completion_probs():
    """
    A quick script to try to estimate how many of them seem to be method completion
    problems
    """
    from datasets import load_dataset

    dataset_all = load_dataset("mbpp")
    lang_spec = PythonLangSpec()
    lang = lang_spec.get_tree_sitter_lang()
    num_single_func = 0
    num_has_imports_single_func = 0
    num_total = 0
    num_has_imports = 0
    for ex in _iterate_for_split(dataset_all, MbppSplit.ALL):
        tree = lang_spec.get_tree_sitter_tree(ex["code"])
        # find all the function definitions
        query = lang.query("(function_definition name: (identifier) @function)")
        captures = query.captures(tree.root_node)
        len(list(captures))
        def_count = ex["code"].count("def ")
        class_count = ex["code"].count("class ")
        import_count = ex["code"].count("import ")
        has_imports = import_count > 0
        if def_count == 1 and class_count == 0:
            # for line in ex["code"].splitlines():
            #    if line.startswith("def "):
            #        assert line.strip().endswith(":")
            num_single_func += 1
            if has_imports:
                num_has_imports_single_func += 1
        num_total += 1
        if has_imports:
            num_has_imports += 1
    print("num_total", num_total)
    print("num_single_func", num_single_func)
    print("not single func", num_total - num_single_func)
    print("num_has_imports_single_func", num_has_imports_single_func)
    print("num_has_imports", num_has_imports)
    print(len(_task_ids_where_reference_solution_fails()))


if __name__ == "__main__":
    # print_problems_with_failing_reference_solutions()
    _measure_method_completion_probs()



import logging
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable
from dataclasses import dataclass
from typing import Any

from tree_sitter import Node, Tree
#from tree_sitter_languages import get_language, get_parser
from tree_sitter_language_pack import get_language, get_parser
#import tree_sitter_python

from synthegrator.lang_specs.lang_util import TreeSitterUtil


class LangSpec(ABC):
    _cls_tree_sitter_parser = None

    def __eq__(self, other):
        if not isinstance(other, LangSpec):
            return NotImplemented
        result = type(self) is type(other)
        return result

    @abstractmethod
    def truncate_function_completion(self, completion: str) -> str: ...

    def get_space_indent_size(self) -> int:
        return 4

    def is_expression_compatable_with(self, other: "LangSpec") -> bool:
        return type(self) == type(other)

    def get_method_name(self, method_src) -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def split_before_last_method_name(prompt: str) -> tuple[str, str]:
        """Splits right before the last method declaration in a string."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_default_docker_env() -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_lang_md_name() -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_tree_sitter_lang_name() -> str:
        raise NotImplementedError

    # Treesitter-based manipulation
    @classmethod
    def get_tree_sitter_parser(cls) -> str:
        if cls._cls_tree_sitter_parser is None:
            cls._cls_tree_sitter_parser = get_parser(
                cls.get_tree_sitter_lang_name(),
            )
        return cls._cls_tree_sitter_parser

    @classmethod
    def get_tree_sitter_tree(cls, src) -> Tree:
        parser = get_parser(cls.get_tree_sitter_lang_name())
        result: Tree = parser.parse(bytes(src, "utf8"))
        return result

    @classmethod
    @abstractmethod
    def get_prefix_comments(cls, src: str | Tree) -> Iterable["PrefixComment"]:
        raise NotImplementedError

    def get_syntax_errors(self, content: str) -> list["SyntaxError"]:
        tree = self.get_tree_sitter_tree(content)
        errors = []

        def traverse(node):
            if node.has_error:
                # Get the start position of the node
                start_line, start_column = node.start_point
                errors.append(SyntaxError("Syntax error", start_line, start_column))
            for child in node.children:
                traverse(child)

        traverse(tree.root_node)
        return errors

    def check_no_syntax_errors(self, content: str) -> bool:
        return len(self.get_syntax_errors(content)) == 0

    def dict_serialize(self) -> dict[str, Any]:
        return {
            "lang_md_name": self.get_lang_md_name(),
        }

    @classmethod
    def get_comment_line_start(cls):
        """
        A symbol that represents a comment line. It may not
        be the only way to start a comment line in the language,
        but is one valid way.
        """
        return "#"


@dataclass(frozen=True)
class SyntaxError:
    message: str
    line_idx: int | None
    column_idx: int | None


def lang_spec_for_path(path: str) -> LangSpec:
    # TODO: This should be specific for the problem
    if path.endswith(".py"):
        from synthegrator.lang_specs.lang_spec_python import PythonLangSpec

        return PythonLangSpec()
    return None


class PythonFunctionParser:
    # TODO: Merge with `Function` class.
    def __init__(self, function_source: str) -> None:
        self.language = get_language("python")
        self.parser = get_parser("python")
        self.function_source = function_source
        self.function_without_doc = None
        self.function_definition_with_doc = None
        self.function_definition_without_doc = None
        self.body_span_begin = -1
        self.body_span_end = -1
        self.has_docstring = False
        self.docstring_node: Node | None = None
        self.parse_source()

    def parse_source(self) -> Node:
        tree = self.parser.parse(bytes(self.function_source, "utf8"))
        root_node = tree.root_node

        classes = [
            node for node in tree.root_node.children if node.type == "class_definition"
        ]
        if classes:
            msg = "Class was defined at root! This class handles functions only."
            raise Exception(
                msg,
            )

        for idx, function_definition in enumerate(
            self.get_function_definitions(root_node),
        ):
            if idx > 0:
                msg = "Second function encountered!"
                raise Exception(msg)
            function_body = self.get_last_block(function_definition)
            docstring_node = self.get_docstring_node(function_body)
            if docstring_node is not None:
                self.has_docstring = True
                self.docstring_node = docstring_node

            self.function_without_doc = TreeSitterUtil.copy_tree_excluding_node(
                root_node,
                docstring_node,
                self.function_source,
            )

            self.function_definition_without_doc = (
                TreeSitterUtil.copy_tree_excluding_node(
                    root_node,
                    function_body,
                    self.function_source,
                )
            )

            self.body_expressions = self.get_non_docstring_nodes(function_body)

            if len(self.body_expressions) == 0:
                self.body_span_begin = docstring_node.end_point[0]
                self.body_span_end = docstring_node.end_point[0]
            else:
                self.body_span_begin = self.body_expressions[0].start_point[0]
                self.body_span_end = self.body_expressions[-1].end_point[0]

            self.function_definition_with_doc = (
                TreeSitterUtil.copy_tree_excluding_nodes(
                    root_node,
                    self.body_expressions,
                    self.function_source,
                ).rstrip()
            )

    def print_node_source(self, node: Node, source_code: str):
        if node.type:  # Check if the node is named (has a type)
            # Print the type of the node and its corresponding source code
            logging.info(source_code[node.start_byte : node.end_byte])

            # Iterate through the children and call the function recursively
            for child in node.children:
                self.print_node_source(child, source_code)

    def get_docstring_node(self, function_node) -> Node:
        first_node = function_node.children[0]
        if (
            first_node.type == "expression_statement"
            and first_node.children[0].type == "string"
        ):
            return first_node
        return None

    @staticmethod
    def get_function_definitions(node: Node) -> Generator[Node, Any, None]:
        if node.type == "function_definition":
            yield node
        for child in node.children:
            if child.type == "function_definition":
                yield child
            elif child.type == "decorated_definition":
                for c in child.children:
                    if c.type == "function_definition":
                        yield c

    def get_non_docstring_nodes(self, function_block: Node) -> list[Node]:
        docstring = self.get_docstring_node(function_block)
        return [
            node
            for node in function_block.children
            if not TreeSitterUtil.nodes_are_equal(node, docstring)
        ]

    def get_last_block(self, node: Node) -> Node:
        for i_node in reversed(node.children):
            if i_node.type == "block":
                return i_node
        msg = "No block found. Empty function?"
        raise Exception(msg)

    def copy_tree_excluding_node(
        self,
        node,
        skip: Node,
        source_code: str,
    ) -> str:
        return self.copy_tree_excluding_nodes(node, [skip], source_code)

    def copy_tree_excluding_nodes(
        self,
        node: Node,
        skip: list[Node],
        source_code: str,
    ) -> str:
        new_source_code = ""

        # Iterate through the children and recursively call the function.
        child_index = 0
        skipped = False
        for child in node.children:
            if any(TreeSitterUtil.nodes_are_equal(child, _skip) for _skip in skip):
                skipped = True
            else:
                # Add the source code between this child and the previous child (or the start of the parent).
                new_source_code += source_code[
                    node.start_byte + child_index : child.start_byte
                ]

                # Recursively handle the child.
                new_source_code += self.copy_tree_excluding_nodes(
                    child,
                    skip,
                    source_code,
                )

            # Update the index to the end of this child.
            child_index = child.end_byte - node.start_byte

        # Add the source code after the last child.
        if skipped:
            new_source_code = new_source_code.lstrip()

        new_source_code += source_code[node.start_byte + child_index : node.end_byte]

        return new_source_code


@dataclass(frozen=True)
class LsFunctionDef:
    """
    A wrapper around a function definition with useful functions common
    between languages.
    """

    node: Node  # A node of type `function_definition`
    tree: Tree  # The tree the node came from

    def get_body_node(self) -> Node:
        """Returns the node that corresponds to the body of the function"""
        raise NotImplementedError

    # @frozen_method_cache
    def get_body_src(
        self,
        include_body_indent: bool = True,
        include_trailing_new_line: bool = True,
        include_prefix_comment: bool = True,
    ) -> str:
        """
        Source of the method (potentially including docstring). The body node
        won't include the body indent (the leading whitespace). Instead that is part
        of the function declaration. The `include_body_indent` can be used to grab it
        
        :param include_prefix_comment: If False, excludes the prefix comment (docstring in Python) from the body
        """
        raise NotImplementedError

    def get_preceding_src_code(self):
        """Returns the source code before the function declaration"""
        raise NotImplementedError

    def get_src_code_after(self):
        """Returns the source code after the function"""
        raise NotImplementedError

    # @frozen_method_cache
    def get_function_name(self) -> str:
        """Returns the name identifier of the function"""
        raise NotImplementedError

    # @frozen_method_cache
    def get_declaration_str(self, include_body_indent: bool = False) -> str:
        r"""
        Returns the full declaration of the function (so the `def`, signature, args, and `:`

        :param include_body_indent: If false the return might have leading ident.
            For example "def foo():\n    pass" would return "def foo():\n    "
            with the indent. If false would return "def foo():\n"
        """
        raise NotImplementedError

    # @frozen_method_cache
    def get_full_function_src(self, zero_base_indent: bool = False) -> str:
        """Returns the entire source code of the function, optionally without leading indentation"""
        raise NotImplementedError

    # @frozen_method_cache
    def get_prefix_comment_str(self, as_plain_text: bool = False) -> str | None:
        """
        Gets the prefix comment (a docstring in Python)

        :param as_plain_text: Normally will include the language specific annotations
            (such as the triple quotes in Python). If true, will return the dedented plain text.
        """
        raise NotImplementedError


@dataclass(slots=True, frozen=True)
class PrefixComment:
    doc_string_node: Node
    function_node: Node

    @property
    def doc_string_str(self) -> str:
        return self.doc_string_node.text.decode()



import threading
import itertools
import traceback
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor

from tqdm import tqdm

from synthegrator.code_problems import CodeProblem, CodeSolution
from synthegrator.code_solver import BaseCodeSolver
from synthegrator.solution_eval import SolutionEvaluation, SolutionEvaluator


class SolveThread(threading.Thread):
    def __init__(
        self,
        solver: BaseCodeSolver,
        problem: CodeProblem,
    ) -> None:
        super().__init__()
        self.solver = solver
        self.problem = problem

    def run(self) -> CodeSolution | None:
        try:
            solution = self.solver.solve(self.problem)
        except (
            Exception
        ) as e:  # We don't want one thread failure to hang the experiment
            print("Solver exception")
            print(e)
            print(traceback.format_exc())
        else:
            return solution
        return None


class EvaluateThread(threading.Thread):
    def __init__(
        self,
        solution: CodeSolution,
        evaluator: SolutionEvaluator = None,
    ) -> None:
        super().__init__()
        self.solution = solution
        self.evaluator = evaluator
        if solution is not None and evaluator is None:
            self.evaluator = solution.problem.preferred_solution_evaluator

    def run(self) -> SolutionEvaluation:
        try:
            if self.solution is None:
                msg = "Solution is None"
                raise ValueError(msg)
            evaluation = self.evaluator.evaluate(self.solution)
        except (
            Exception
        ) as e:  # We don't want one thread failure to hang the experiment
            if "Too many open files" in str(e):
                # Custom error message
                custom_message = (
                    f"{e!s}\n\nThis is a previously observed known issue. "
                    "It might either be a bug/leak in synthegrator, or if you are "
                    "actually running on a massive "
                    "number of threads it might be a ulimit problem where it can be fixed by "
                    "increasing limits. The exact recommendation is not clear yet though. "
                    "Please make a GitHub issue if you observe."
                )
                raise Exception(custom_message) from e
            return SolutionEvaluation(
                solution=self.solution,
                test_results=None,
                extra_metrics=[],
                main_metric=None,
                exception=e,
                exception_traceback=traceback.format_exc(),
            )
        else:
            return evaluation


def solve_all_problems(
    solver: BaseCodeSolver,
    problems: Iterable[CodeProblem],
    max_threads: int = 1,
    cuda_clear_freq: int = 100,
) -> Iterable[CodeSolution]:
    """
    Solve all problems in an iterable.

    Args:
    ----
        solver (BaseCodeSolver): The solver to use.
        problems (Iterable[CodeProblem]): The problems to solve. You might
            want to slice it beforehand if you want to limit the number of
            problems solved.
        max_threads (int, optional): The maximum number of threads to use.
            Defaults to 1 (no multi-threading). You should also check whether
            solver.allows_multithreading is True.
        cuda_clear_freq (int, optional): The frequency at which to clear the
            CUDA cache. Defaults to 100 assuming cuda is available (else will
            not do anything).

    """
    # Safety check for None or empty problems
    if problems is None:
        raise ValueError("problems parameter cannot be None")

    is_empty, problems = is_generator_empty(problems)
    if is_empty:
        return iter(())
    
    # Proceed with the rest of the function
    try:
        import torch
    except ImportError:
        cuda_clear_freq = None
    else:
        if not torch.cuda.is_available():
            cuda_clear_freq = None
    if max_threads > 1 and not solver.allows_multithreading:
        msg = f"{solver.__class__.__name__} does not allow multithreading"
        raise ValueError(
            msg,
        )
    if cuda_clear_freq and max_threads > 1:
        msg = "cuda_clear_freq and max_threads > 1 are incompatible"
        raise ValueError(
            msg,
        )
    cuda_reset_counter = 0

    solve_threads = (
        SolveThread(
            solver,
            problem,
        )
        for problem in problems
    )

    def pbar(data):
        return tqdm(
            data,
            total=len(problems) if hasattr(problems, "__len__") else None,
            desc="Solving problems",
        )

    if max_threads <= 1:
        for solve_thread in pbar(
            solve_threads,
        ):
            if cuda_clear_freq is not None and cuda_reset_counter > cuda_clear_freq:
                torch.cuda.empty_cache()
                cuda_reset_counter = 0
            solution = solve_thread.run()
            if solution is not None:
                yield solution
            cuda_reset_counter += 1
    else:
        with ThreadPoolExecutor(max_workers=max_threads) as service:
            futures = [
                service.submit(
                    thread.run,
                )
                for thread in solve_threads
            ]
            for future in pbar(
                futures,
            ):
                yield future.result(timeout=259200)

            service.shutdown()


def is_generator_empty(generator: Iterable) -> tuple[bool, Iterable]:
    """Check if a generator is empty and return a usable copy.
    
    Returns:
        A tuple of (is_empty, usable_generator)
        - is_empty: True if the generator is empty, False otherwise
        - usable_generator: A generator that includes all original items,
          including the first one checked for emptiness
    """
    # For sequences with a length, just check the length
    if hasattr(generator, "__len__"):
        return len(generator) == 0, generator
    
    # For generators, use tee to make copies
    try:
        # Create two copies - one to check for emptiness, one to return
        check_copy, return_copy = itertools.tee(generator)
        first_item = next(check_copy, None)
        
        if first_item is None:
            # Generator is empty
            return True, return_copy
        else:
            # Generator has items - create a chain with the first item and the rest
            return False, itertools.chain([first_item], check_copy)
    except Exception:
        # If we can't check, assume it's not empty to be safe
        return False, generator


def evaluate_all_solutions(
    solutions: Iterable[CodeSolution],
    evaluator: SolutionEvaluator = None,
    max_threads: int = 1,
) -> Iterable[SolutionEvaluation]:
    """
    Evaluate all solutions in an iterable.

    Args:
    ----
        solutions (Iterable[CodeSolution]): The solutions to evaluate.
        evaluator (SolutionEvaluator): The evaluator to use. If it is None,
            it will use the preferred evaluator of the problem.
        max_threads (int, optional): The maximum number of threads to use.
            Defaults to 1 (no multi-threading).

    """
    # Safety check for solutions
    if solutions is None:
        raise ValueError("solutions parameter cannot be None")
        
    is_empty, solutions = is_generator_empty(solutions)
    if is_empty:
        return iter(())

    evaluate_threads = (
        EvaluateThread(
            solution,
            evaluator,
        )
        for solution in solutions
    )

    def pbar(data):
        return tqdm(
            data,
            total=len(solutions) if hasattr(solutions, "__len__") else None,
            desc="Evaluating solutions",
        )

    if max_threads <= 1:
        for evaluate_thread in pbar(
            evaluate_threads,
        ):
            evaluation = evaluate_thread.run()
            yield evaluation
    else:
        with ThreadPoolExecutor(max_workers=max_threads) as service:
            futures = [
                service.submit(
                    thread.run,
                )
                for thread in evaluate_threads
            ]
            for future in pbar(
                futures,
            ):
                yield future.result(timeout=259200)

            service.shutdown()


def solve_and_evaluate_problems(
    solver: BaseCodeSolver,
    problems: Iterable[CodeProblem],
    max_threads_solve: int = 1,
    max_threads_eval: int = 1,
    finish_all_solves_first: bool = True,
    evaluator: SolutionEvaluator = None,
) -> Iterable[SolutionEvaluation]:
    # Safety check for problems
    if problems is None:
        raise ValueError("problems parameter cannot be None")

    solves = solve_all_problems(
        solver,
        problems,
        max_threads=max_threads_solve if solver.allows_multithreading else 1,
        cuda_clear_freq=None if solver.allows_multithreading else 100,
    )
    
    # Safety check for solves
    if solves is None:
        raise ValueError("solve_all_problems returned None instead of an iterable")

    if finish_all_solves_first:
        try:
            solves = list(solves)
        except TypeError as e:
            if "not iterable" in str(e):
                raise ValueError(f"solves is not iterable: {e}") from e
            raise

        for solve in solves:
            if solve is not None:
                if solve.problem is None:
                    raise RuntimeError("Problem is None")
    return evaluate_all_solutions(solves, evaluator=evaluator, max_threads=max_threads_eval)



"""
Various utilities for rendering code problems into prompt strings.
These are intended as fairly standard versions. Certain models might
choose to do some form of specialized conversions.
"""

import dataclasses
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable

from lmwrapper.abstract_predictor import LmPredictor
from lmwrapper.structs import ChatGptRoles, LmChatDialog, LmChatTurn, LmPrompt

from synthegrator.code_problems import (
    CodeProblem,
    TestCase,
    TestCaseMethodCallIsEq,
)
from synthegrator.few_shotting import FewShotConfig
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.prompting_test_case_selection import (
    PromptingTestCaseSelectionStrategy,
)
from synthegrator.transformation_spec import (
    MarkElement,
    MarkText,
    StsPlaceTransforming,
    get_verbs_per_path,
    markup_path,
)
from synthegrator.util import ensure_newline_at_end


@dataclasses.dataclass(frozen=True)
class LmPromptRender:
    prompt: LmPrompt


class ProblemNotSupportedError(ValueError):
    """
    Raised by a PromptRender when the .render is given
    a problem that the renderer does not support.
    """


@dataclasses.dataclass(frozen=True)
class LmPromptRenderSingleEdit(LmPromptRender):
    code_prefix: str
    """The code prefix is the source code given to the model
    that is a prefix to any output solution. This is useful because
    sometimes dialogue models will repeat part of the prompt in the
    output. Thus can be used extract out just the new contents"""
    path: str
    """The path of the file that is being edited. Needed when making
    the solve step."""
    changed_element: MarkElement[StsPlaceTransforming] | None
    """The element that is changed in by this edit. Needed when making the
     solve step."""


@dataclasses.dataclass(frozen=True)
class SingleLineBugRender(LmPromptRender):
    code_prefix: str
    """The code prefix is the source code given to the model
    that is a prefix to any output solution. This is useful because
    sometimes dialogue models will repeat part of the prompt in the
    output. Thus can be used extract out just the new contents"""
    path: str
    """The path of the file that is being edited. Needed when making
    the solve step."""
    changed_element: MarkElement[StsPlaceTransforming] | None
    """The element that is changed in by this edit. Needed when making the
     solve step."""


class PromptRenderer(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmPromptRender:
        """
        Renders a code problem into a prompt for a language model.
        :param problem: The code problem to render
        :param lm: The language model to render for. Useful if want to
            determine whether to make a dialog or how long can generate.
        :param prompt_seed: A seed to use for any randomization (e.g. for
             few shot seed.)
        """
        raise NotImplementedError

    def __call__(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmPromptRender:
        return self.render(problem, lm, prompt_seed)


class PromptRendererSingleEdit(PromptRenderer):
    def __init__(self):
        super().__init__()

    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmPromptRenderSingleEdit:
        raise NotImplementedError

    def __call__(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmPromptRenderSingleEdit:
        return self.render(problem, lm, prompt_seed)


RenderFuncSingleEdit = Callable[
    [CodeProblem, LmPredictor, int | None],
    LmPromptRenderSingleEdit,
]


def render_markup_to_text_autoregressive(
    markup: MarkElement,
    path: str,
    problem: CodeProblem,
    include_instructions: bool | None = None,
) -> LmPromptRenderSingleEdit:
    """Renders markup up to the first edit/insert statement. Masks any invisible text."""
    text = []
    final_instructions = ""
    if include_instructions is None:
        include_instructions = (
            problem.instructions_are_essential and problem.instructions
        )
    if include_instructions:
        lang_spec = problem.get_lang_spec_for_path(path)
        final_instructions = (
            # fmt: off
            f"{lang_spec.get_comment_line_start() * 2} Instructions\n"
            + f"{lang_spec.get_comment_line_start()} ".join(
                problem.instructions.splitlines(keepends=True)
            )
            + "\n"
            + f"{lang_spec.get_comment_line_start() * 2}\n"
            # fmt: on
        )

    def _finish_up(node: MarkElement[StsPlaceTransforming]):
        joined_text = "".join(text)
        stop = []
        if node is not None and node.verb.max_length_lines == 1:
            stop = ["\n"]
        if node is not None and node.verb.stop_at_block_end:
            # Extra stops like in https://github.com/bigcode-project/bigcode-evaluation-harness/blob/136b93c0aea/lm_eval/tasks/humaneval.py#L54
            # TODO multilingual
            if node.verb.lang_spec_name == "python":
                # TODO make it work when not starting at the 0 block
                # TODO move this to autoregressive renderer
                stop.extend(["\nclass", "\ndef", "\n#"])
        if len(stop) == 0:
            stop = None

        prompt = LmPrompt(
            final_instructions + joined_text,
            stop=stop,
        )
        return LmPromptRenderSingleEdit(
            prompt,
            joined_text,
            path,
            node,
        )

    for state in markup.iterate_parser_states():
        if isinstance(state.node, MarkText):
            if state.is_visible:
                text.append(state.node.text)
        elif isinstance(state.node, MarkElement) and state.is_editable:
            return _finish_up(state.node)
    return _finish_up(None)


def render_markup_to_plain_text(
    markup: MarkElement,
) -> str:
    text = []
    for state in markup.iterate_parser_states():
        if isinstance(state.node, MarkText) and state.is_visible:
            text.append(state.node.text)
    return "".join(text)


def render_single_edit_codeproblem_autoregressive(
    problem: CodeProblem,
    prompt_seed: int | None = None,
) -> LmPromptRenderSingleEdit:
    """
    Basic rendering (as in converting to a string prompt) of a code problem as
    a single file autoregressive prompt.
    Only works if there is a single edit or insertion point in the TransformationSpec
    """
    spec = problem.transformation_spec
    if spec.count_editing_statements() != 1:
        msg = "Currently simple and assumes a exactly one edit statement"
        raise NotImplementedError(
            msg,
        )
    path_to_verbs = get_verbs_per_path(problem.working_directory, spec)
    if len(path_to_verbs) != 1:
        msg = "Currently simple and assumes a single path"
        raise ValueError(msg)

    path, _verbs = path_to_verbs.popitem()
    markedup_text = markup_path(
        problem.working_directory,
        path,
        problem.transformation_spec,
    )
    out = render_markup_to_text_autoregressive(
        markedup_text,
        path,
        problem,
    )
    if out.changed_element is None:
        msg = "No insert element found in the markup"
        raise RuntimeError(msg)
    return out


def render_single_edit_code_repair_problem_autoregressive(
    problem: CodeProblem,
    prompt_seed: int | None = None,
    prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy | None = None,
) -> LmPromptRenderSingleEdit:
    """Renders markup up to the first edit/insert statement. Masks any invisible text."""
    if prompt_test_case_selection_strategy is not None:
        msg = "Currently not implemented for this renderer"
        raise NotImplementedError(
            msg,
        )
    spec = problem.transformation_spec
    if spec.count_editing_statements() != 1:
        msg = "Currently simple and assumes a exactly one edit statement"
        raise NotImplementedError(
            msg,
        )
    path_to_verbs = get_verbs_per_path(problem.working_directory, spec)
    if len(path_to_verbs) != 1:
        msg = "Currently simple and assumes a single path"
        raise ValueError(msg)

    path, _verbs = path_to_verbs.popitem()
    markedup_text = markup_path(
        problem.working_directory,
        path,
        problem.transformation_spec,
    )

    text = []
    final_instructions = (
        "\nQuestion: There is a bug in the above code snippet tagged by <buggy> and"
        " </buggy>. Please generate the correct version.\nAnswer:\n<fixed>"
    )

    def _finish_up(node: MarkElement[StsPlaceTransforming]):
        joined_text = "".join(text)
        stop = None
        if node is not None:
            stop = ["</fixed>"]
        prompt = LmPrompt(
            joined_text + final_instructions,
            stop=stop,
            cache=False,
        )
        return SingleLineBugRender(
            prompt,
            joined_text,
            path,
            node,
        )

    edit_node = None
    for state in markedup_text.iterate_parser_states():
        if isinstance(state.node, MarkText):
            if state.is_visible:
                text.append(state.node.text)
        elif isinstance(state.node, MarkElement) and state.is_editable:
            if edit_node is not None:
                msg = "Currently simple and assumes a exactly one edit statement"
                raise NotImplementedError(
                    msg,
                )
            edit_node = state.node
            text.append("<buggy>")

            def after_done():
                text[-1] = text[-1].rstrip()
                text.append("</buggy>\n")

            state.after_visit_children_callbacks.append(after_done)

    out = _finish_up(edit_node)

    if out.changed_element is None:
        msg = "No insert element found in the markup"
        raise RuntimeError(msg)
    return out


def render_single_edit_codeproblem_dialog(
    problem: CodeProblem,
    few_shot_config: FewShotConfig = None,
    prompt_seed: int | None = None,
    prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy = None,
) -> LmPromptRenderSingleEdit:
    messages = []

    def select_prompt_tc(p):
        return (
            prompt_test_case_selection_strategy.select_for_problem(p)
            if prompt_test_case_selection_strategy is not None
            else None
        )

    if few_shot_config is not None:
        for few_shot in few_shot_config.library.sample_problems(
            n=few_shot_config.num_examples,
            target_problem=problem,
            seed=prompt_seed,
        ):
            new_messages, auto_regressive = _problem_instance_dialog(
                few_shot,
                ex_index=len(messages),
                add_answer=True,
                prompt_test_cases=select_prompt_tc(few_shot),
            )
            messages.extend(new_messages)
    # add final prompt
    new_messages, auto_regressive = _problem_instance_dialog(
        problem,
        ex_index=len(messages),
        add_answer=False,
        prompt_test_cases=select_prompt_tc(problem),
    )
    messages.extend(new_messages)
    return LmPromptRenderSingleEdit(
        prompt=LmPrompt(
            LmChatDialog(
                values=messages,
            ),
        ),
        code_prefix=auto_regressive.prompt.text,
        path=auto_regressive.path,
        changed_element=auto_regressive.changed_element,
    )


def format_together_test_cases_test_suite(
    test_cases: list[TestCase],
    ex_index: int = 0,
) -> str:
    if test_cases is None:
        return ""
    msg = []
    py_spec = PythonLangSpec()
    for i, test_case in enumerate(test_cases):
        category_name = f"ex{ex_index}"
        if isinstance(test_case, TestCaseMethodCallIsEq):
            category_name = f"{test_case.method_name}"
        override_name = f"{category_name}_{i}"
        msg.append(
            test_case.format_for_test_framework(
                py_spec,
                override_test_id=override_name,
            ),
        )
    return "\n".join(msg)


def _problem_instance_dialog(
    problem: CodeProblem,
    ex_index: int,
    add_answer: bool,
    prompt_test_cases: list[TestCase] | None = None,
) -> tuple[list[LmChatTurn], LmPromptRenderSingleEdit]:
    auto_regressive = render_single_edit_codeproblem_autoregressive(problem)
    out = []
    instructions = problem.instructions
    verb = auto_regressive.changed_element.verb
    if instructions is None or instructions == "":
        # A default instructions to get the chat model like a completion model
        instructions = (
            "Give a completion to the following code to the best "
            "of your ability as if acting like a smart completion model. "
            "Give only the completion, no other text. "
            "Do not put in a markdown code block. "
        )
        if verb.max_length_lines == 1:
            instructions += "For now we are only predicting the next line. "
            instructions += "Give the text of the next line."
        else:
            instructions += "Do not include any other text than the completion."

    message = f"{instructions}\n" + "```\n" + f"{auto_regressive.prompt.text}"
    message = ensure_newline_at_end(message)
    message += "```\n"
    tests_text = format_together_test_cases_test_suite(
        prompt_test_cases,
        ex_index,
    )
    if tests_text:
        message += "Test cases:\n```\n"
        message += ensure_newline_at_end(tests_text)
        message += "```"
    prompt_msg = LmChatTurn(
        role=ChatGptRoles.user,
        content=message,
    )
    out.append(prompt_msg)
    if add_answer:
        if len(problem.known_solutions) != 1:
            msg = (
                "In few shot currently only support problems with exactly one solution"
            )
            raise NotImplementedError(
                msg,
            )
        steps = problem.known_solutions[0].solve_steps
        if len(steps) != 1:
            msg = "In fewshot currently only supports single step solutions"
            raise ValueError(msg)
        answer_content = f"```\n{steps[0].value}"
        if not answer_content.endswith("\n"):
            answer_content += "\n"
        answer_content += "```"
        out.append(
            LmChatTurn(
                role=ChatGptRoles.assistant,
                content=answer_content,
            ),
        )
    return out, auto_regressive


class PromptRendererSingleEditAutoregressive(PromptRendererSingleEdit):
    def __init__(
        self,
        few_shot_config: FewShotConfig = None,
        prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy = None,
    ):
        super().__init__()
        self._few_shot_config = few_shot_config
        self._prompt_test_case_selection_strategy = prompt_test_case_selection_strategy

    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmPromptRenderSingleEdit:
        if self._few_shot_config is not None:
            warnings.warn(
                "Few shot not supported with this non-chat version. "
                "No few shot will be used. TODO",
            )
        return render_single_edit_codeproblem_autoregressive(problem, prompt_seed)


class PromptRendererSingleEditDialog(PromptRendererSingleEdit):
    def __init__(
        self,
        few_shot_config: FewShotConfig = None,
        prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy = None,
    ):
        super().__init__()
        self._few_shot_config = few_shot_config
        self._prompt_test_case_selection_strategy = prompt_test_case_selection_strategy

    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmPromptRenderSingleEdit:
        return render_single_edit_codeproblem_dialog(
            problem,
            few_shot_config=self._few_shot_config,
            prompt_seed=prompt_seed,
            prompt_test_case_selection_strategy=self._prompt_test_case_selection_strategy,
        )


class PromptRendererSingleEditGeneric(PromptRendererSingleEdit):
    """
    A prompt render that combines several prompt renderers and tries to choose
    the best one based on the problem and language model.
    """

    @classmethod
    def build_from_defaults(
        cls,
        max_tokens: int = 500,
        force_use_dialogue: bool = False,
        few_shot_config: FewShotConfig = None,
        prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy = None,
    ) -> "PromptRendererSingleEditGeneric":
        """
        The different defaults component renderers have some common components (such
        as the few_shot_config). This builder method will construct all the defaults
        with the supplied configs.
        """
        return cls(
            max_tokens=max_tokens,
            force_use_dialogue=force_use_dialogue,
            completion_render_func=cls.default_completion(
                few_shot_config=few_shot_config,
                prompt_test_case_selection_strategy=prompt_test_case_selection_strategy,
            ),
            dialogue_render_func=cls.default_dialogue(
                few_shot_config,
                prompt_test_case_selection_strategy,
            ),
        )

    @classmethod
    def default_completion(
        cls,
        few_shot_config: FewShotConfig = None,
        prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy = None,
    ):
        return PromptRendererSingleEditAutoregressive(
            few_shot_config=few_shot_config,
            prompt_test_case_selection_strategy=prompt_test_case_selection_strategy,
        )

    @classmethod
    def default_dialogue(
        cls,
        few_shot_config: FewShotConfig = None,
        prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy = None,
    ):
        return PromptRendererSingleEditDialog(
            few_shot_config=few_shot_config,
            prompt_test_case_selection_strategy=prompt_test_case_selection_strategy,
        )

    def __init__(
        self,
        max_tokens: int = 500,
        force_use_dialogue: bool = False,
        completion_render_func: PromptRendererSingleEdit = None,
        dialogue_render_func: PromptRendererSingleEdit = None,
        few_shot_config_default: FewShotConfig = None,
    ):
        super().__init__()
        self._max_tokens = max_tokens
        self._force_use_dialogue = force_use_dialogue
        self._completion_render_func = (
            completion_render_func
            or self.default_completion(few_shot_config=few_shot_config_default)
        )
        self._dialogue_render_func = dialogue_render_func or self.default_dialogue(
            few_shot_config=few_shot_config_default,
        )

    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmPromptRenderSingleEdit:
        if not lm.is_chat_model or self._force_use_dialogue:
            render = self._completion_render_func(problem, lm, prompt_seed)
        else:
            render = self._dialogue_render_func(problem, lm, prompt_seed)
        dataclasses.replace(
            render,
            prompt=dataclasses.replace(
                render.prompt,
                max_tokens=self._max_tokens,
                cache=True,
            ),
        )
        return render



import itertools
import logging
import os
from itertools import islice
from unittest.mock import MagicMock

import pytest
from tqdm.auto import tqdm

from synthegrator.code_solver import (
    LmCodeSolverAutoRegressive,
)
from synthegrator.execution_threading import evaluate_all_solutions, solve_all_problems
from synthegrator.memory_fs import ProjectDir
from synthegrator.problem_rendering_insertion_tags import (
    TaggedEditRenderer,
    TaggedEditResponseParser,
)
from synthegrator.sandboxing import run_on_docker
from synthegrator.solution_eval import (
    evaluate_code_problem_execution,
    parse_junit_test_cases,
)
from synthegrator.synthdatasets.defects4j import (
    DEFECTS_4_J_DOCKER_ENV,
    _default_projects_root,
    _get_deepest_path,
    make_discovered_test_case,
    yield_defects4j,
)

logging.getLogger().setLevel(level="DEBUG")


IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"
SKIP_DOCKER_TESTS = True
skip_docker_reason = "Docker install takes a long time"


def test_read_from_zip():
    from synthegrator.synthdatasets.defects4j_data.dataset import read_file_from_zip

    content = read_file_from_zip(
        "Defects4J_projects_clean/Chart_1/source/org/jfree/chart/renderer/category/AbstractCategoryItemRenderer.java"
    )
    assert content.startswith(
        "/* ===========================================================\r\n"
        " * JFreeChart : a free chart library for the Java(tm) platform"
    )


@pytest.mark.skipif(SKIP_DOCKER_TESTS, reason=skip_docker_reason)
@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Downloads large docker images")
def test_can_run_work_passing():
    project_name = "Chart"
    problem_id = "1"
    start_path = _default_projects_root() / f"{project_name}_{problem_id}"
    file_path = str(
        _get_deepest_path(
            start_path,
            file_only=True,
        ).relative_to(start_path),
    )
    assert (
        file_path
        == "source/org/jfree/chart/renderer/category/AbstractCategoryItemRenderer.java"
    )
    assert (start_path / file_path).exists()
    discovered_suite = make_discovered_test_case(
        project_name="Chart",
        problem_id="1",
        is_buggy=False,
        file_path=file_path,
    )
    cmd_outputs = run_on_docker(
        docker_context=DEFECTS_4_J_DOCKER_ENV,
        cmds=discovered_suite.cmds,
        interactive_shell_debug=False,
        files=ProjectDir.construct_with_one_file(
            "solution.java",
            _get_deepest_path(
                _default_projects_root() / f"{project_name}_{problem_id}_fixed",
                file_only=True,
            ).read_text(),
        ),
    )
    exec_result = cmd_outputs[-1].stdout.decode()
    test_result = parse_junit_test_cases(exec_result, MagicMock())
    print(test_result)
    assert test_result.all_passed()


@pytest.mark.skipif(SKIP_DOCKER_TESTS, reason=skip_docker_reason)
@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Downloads large docker images")
def test_can_run_work_buggy():
    project_name = "Chart"
    problem_id = "1"
    start_path = _default_projects_root() / f"{project_name}_{problem_id}"
    file_path = str(
        _get_deepest_path(
            start_path,
            file_only=True,
        ).relative_to(start_path),
    )
    assert (
        file_path
        == "source/org/jfree/chart/renderer/category/AbstractCategoryItemRenderer.java"
    )
    assert (start_path / file_path).exists()
    discovered_suite = make_discovered_test_case(
        project_name="Chart",
        problem_id="1",
        is_buggy=True,
        file_path=file_path,
    )
    cmd_outputs = run_on_docker(
        docker_context=DEFECTS_4_J_DOCKER_ENV,
        cmds=discovered_suite.cmds,
        interactive_shell_debug=False,
        files=ProjectDir.construct_with_one_file(
            "solution.java",
            _get_deepest_path(
                _default_projects_root() / f"{project_name}_{problem_id}",
                file_only=True,
            ).read_text(),
        ),
    )
    exec_result = cmd_outputs[-1].stdout.decode()
    test_result = parse_junit_test_cases(exec_result, MagicMock())
    assert not test_result.all_passed()


def test_read_all_problems():
    all_problems = list(yield_defects4j())
    assert len(all_problems) == 119


@pytest.mark.skipif(SKIP_DOCKER_TESTS, reason=skip_docker_reason)
@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Downloads large docker images")
def test_new_sample_eval():
    all_problems = list(yield_defects4j())
    assert len(all_problems) == 119
    for example_defects4j_prob in tqdm(
        itertools.islice(all_problems, 0, None, 20),
        total=119 // 20,
    ):
        print("Problem id:", example_defects4j_prob.problem_id)
        assert len(example_defects4j_prob.known_solutions) == 1
        result = evaluate_code_problem_execution(
            example_defects4j_prob,
            example_defects4j_prob.known_solutions[0],
            try_cache=False,
            do_not_execute_if_syntax_fail=False,
        )
        print("All passed:", result.all_passed())


@pytest.mark.skipif(SKIP_DOCKER_TESTS, reason=skip_docker_reason)
@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Downloads large docker images")
def test_prompt():
    from lmwrapper.openai_wrapper import OpenAiModelNames, get_open_ai_lm

    example_defects4j_prob = next(yield_defects4j())
    print("Problem id:", example_defects4j_prob.problem_id)
    renderer = TaggedEditRenderer()
    lm = get_open_ai_lm(OpenAiModelNames.gpt_4)
    renderer(
        example_defects4j_prob,
        lm,
    )
    solver = LmCodeSolverAutoRegressive(
        lm,
        prompt_renderer=renderer,
        response_parser=TaggedEditResponseParser(),
        include_lm_response=False,
    )
    problems = list(islice(yield_defects4j(), 1))
    solution = solver.solve(problems[0])
    # print("Prompt")
    # print(solution.lm_prediction.prompt)
    # print("Completion text")
    # print(solution.lm_prediction.completion_text)
    assert len(solution.solve_steps) == 1
    # problems = islice(problems, 10)
    thread_goal = 3
    solves = list(
        solve_all_problems(
            solver=solver,
            problems=problems,
            max_threads=thread_goal if solver.allows_multithreading else 1,
            cuda_clear_freq=False,
        ),
    )
    for solve in islice(solves, 10):
        print("Solve value:")
        print(solve.solve_steps[0].value)
    #  print("aplication")
    #  pretty_print_python_code(solve.apply().get_only_file().content_str)
    # exit()
    evals = evaluate_all_solutions(solves, max_threads=thread_goal)
    evals = list(evals)
    print(evals)
    # pass_rate = sum(1 for e in evals if e.main_metric.is_success) / len(evals)
    # print(f"Pass rate: {pass_rate}")


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No api")
def test_prompt2():
    from lmwrapper.openai_wrapper import OpenAiModelNames, get_open_ai_lm

    example_defects4j_prob = next(yield_defects4j())
    print("Problem id:", example_defects4j_prob.problem_id)
    renderer = TaggedEditRenderer()
    # TODO mocked version
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo_instruct)
    render = renderer(
        example_defects4j_prob,
        lm,
    )
    print(render.prompt.text)
    assert render.prompt.text.endswith("Answer:\n" '<edit_solve id="0_1xtCqX">')
    solver = LmCodeSolverAutoRegressive(
        lm,
        prompt_renderer=renderer,
        response_parser=TaggedEditResponseParser(),
        include_lm_response=True,
    )
    problems = list(islice(yield_defects4j(), 1))
    solution = solver.solve(problems[0])
    assert solution.lang_spec.get_lang_md_name() == "java"
    print("Prompt")
    print(solution.lm_prediction.prompt.text)
    print("stop", solution.lm_prediction.prompt.stop)
    print("Completion text")
    print(solution.lm_prediction.completion_text)
    assert len(solution.solve_steps) == 1
    # problems = islice(problems, 10)
    solves = list(
        solve_all_problems(
            solver=solver,
            problems=problems,
            max_threads=1,
            cuda_clear_freq=False,
        ),
    )
    for solve in islice(solves, 10):
        print("Solve value:")
        print(solve.solve_steps[0].value)


@pytest.mark.skipif(SKIP_DOCKER_TESTS, reason=skip_docker_reason)
@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Downloads large docker images")
def test_defect_apply():
    prob = next(yield_defects4j())
    solution = prob.known_solutions[0]
    files = solution.apply()
    print(files["solution.java"].content)
    result = evaluate_code_problem_execution(
        prob,
        prob.known_solutions[0],
        try_cache=False,
        do_not_execute_if_syntax_fail=True,
    )
    assert result.all_passed()
    print("All passed:", result.all_passed())


if __name__ == "__main__":
    pass



import itertools
import logging
from collections import Counter, defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed

import numpy as np
import tqdm
from human_eval.data import HUMAN_EVAL, read_problems, stream_jsonl, write_jsonl

from synthegrator.human_eval.execution import check_correctness


def estimate_pass_at_k(
    num_samples: int | list[int] | np.ndarray,
    num_correct: list[int] | np.ndarray,
    k: int,
) -> np.ndarray:
    """Estimates pass@k of each problem and returns them in an array."""

    def estimator(n: int, c: int, k: int) -> float:
        """Calculates 1 - comb(n - c, k) / comb(n, k)."""
        if n - c < k:
            return 1.0
        return 1.0 - np.prod(1.0 - k / np.arange(n - c + 1, n + 1))

    if isinstance(num_samples, int):
        num_samples_it = itertools.repeat(num_samples, len(num_correct))
    else:
        assert len(num_samples) == len(num_correct)
        num_samples_it = iter(num_samples)

    return np.array(
        [
            estimator(int(n), int(c), k)
            for n, c in zip(num_samples_it, num_correct, strict=True)
        ],
    )


def evaluate_functional_correctness(
    sample_file: str,
    k: list[int] | None = None,
    n_workers: int = 4,
    timeout: float = 3.0,
    problem_file: str = HUMAN_EVAL,
):
    """
    Evaluates the functional correctness of generated samples, and writes
    results to f"{sample_file}_results.jsonl.gz".
    """
    if not k:
        k = [1, 10, 100]

    problems = read_problems(problem_file)

    # Check the generated samples against test suites.
    with ThreadPoolExecutor(max_workers=n_workers) as executor:
        futures = []
        completion_id = Counter()
        n_samples = 0
        results = defaultdict(list)

        logging.info("Reading samples...")
        for sample in tqdm.tqdm(stream_jsonl(sample_file)):
            task_id = sample["task_id"]
            completion = sample["completion"]
            args = (problems[task_id], completion, timeout, completion_id[task_id])
            future = executor.submit(check_correctness, *args)
            futures.append(future)
            completion_id[task_id] += 1
            n_samples += 1

        assert len(completion_id) == len(problems), "Some problems are not attempted."

        logging.info("Running test suites...")
        for future in tqdm.tqdm(as_completed(futures), total=len(futures)):
            result = future.result()
            results[result["task_id"]].append((result["completion_id"], result))

    # Calculate pass@k.
    total, correct = [], []
    for result in results.values():
        result.sort()
        passed = [r[1]["passed"] for r in result]
        total.append(len(passed))
        correct.append(sum(passed))
    total = np.array(total)
    correct = np.array(correct)

    ks = k
    pass_at_k = {
        f"pass@{k}": estimate_pass_at_k(total, correct, k).mean()
        for k in ks
        if (total >= k).all()
    }

    # Finally, save the results in one file:
    def combine_results():
        for sample in stream_jsonl(sample_file):
            task_id = sample["task_id"]
            result = results[task_id].pop(0)
            sample["result"] = result[1]["result"]
            sample["passed"] = result[1]["passed"]
            yield sample

    out_file = sample_file + "_results.jsonl"
    logging.info("Writing results to %s...", out_file)
    write_jsonl(out_file, tqdm.tqdm(combine_results(), total=n_samples))

    return pass_at_k



import ast
import logging
logger = logging.getLogger(__name__)
import math
from collections import defaultdict
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from timeit import default_timer as timer
from typing import Any

import diskcache
from lxml import etree

from synthegrator.code_problems import (
    CodeProblem,
    CodeSolution,
    DiscoveredTestsuite,
    TestCase,
)
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.memory_fs import ProjectDir
from synthegrator.sandboxing import (
    PY_DEFAULT_DOCKER_ENV,
    ExecLimits,
    TestsuiteExecResult,
    pytest_on_docker,
    run_on_docker, CmdExecResult, Cmd,
)
from synthegrator.transformation_spec import (
    MarkElement,
    MarkText,
    SolveStep,
    StsCodeClassificationQuestion,
    StsMarkNode,
    StsPlaceTransforming,
    StsValueQuery,
    markup_path,
)
from synthegrator.util import (
    get_first_line_with_nonwhitespace,
    get_only_item,
    pretty_print_python_code,
)

eval_cache = diskcache.Cache("eval_cache")
# TODO make this user configurable


@dataclass
class TestRunResult:
    success: bool
    output: Any
    completed: bool
    fail_message: str = None
    fail_type: str = None
    fail_text: str = None
    id_: str = None
    name: str = None


@dataclass
class TestRunSummary:
    """Store pytest result summary."""

    __test__ = False
    name: str
    errors: int
    failures: int
    skipped: int
    tests: int
    time: timedelta
    timestamp: datetime
    hostname: str
    valid_xml: bool = True


class TestRunResultList(list[TestRunResult]):
    __test__ = False
    METRIC_NAME: str = "test_passing"

    def __init__(
        self,
        code_problem: CodeProblem,
        summary: TestRunSummary = None,
        initial_list=None,
        meta_error: TestRunResult = None,
        runtime: float | None = None,
        syntax_check_result: bool | None = None,
    ) -> None:
        super().__init__(initial_list or [])

        self.code_problem = code_problem
        self.summary = summary
        self.runtime = runtime
        self.syntax_check_result = syntax_check_result
        self.collection_error = None
        self.exec_error = None
        self.timeout = None


        if meta_error is not None:
            raise NotImplementedError

    def passed(self) -> int:
        if self.summary is None:
            return 0
        return (
            (self.summary.tests - self.summary.skipped) - self.summary.errors
        ) - self.summary.failures

    def all_passed(self) -> bool:
        return len(self) > 0 and all(t.success for t in self)
        if not self.summary:
            return False
        if self.syntax_check_result is False:
            return False
        return self.summary.errors == 0 and self.summary.failures == 0

    def pass_rate(self) -> float:
        if not self.summary:
            return 0.0
        if not self.syntax_check_result and (
            self.summary.errors is None or self.summary.tests is None
        ):
            return 0.0
        return self.passed() / max(
            (self.summary.tests - self.summary.skipped),
            1,
        )

    def as_metric(self) -> "SolutionMetric":
        return SolutionMetric(
            is_success=self.all_passed(),
            float_val=self.pass_rate(),
            label=self.METRIC_NAME,
        )

    def __str__(self) -> str:
        if not self.summary:
            if self.syntax_check_result is False:
                return "Syntax check failed!"
            return "No summary available!"
        table = "{:<15} | {:<15}\n".format("Attribute", "Value")  # header
        table += "{:<15} | {:<15}\n".format(
            "-" * 15,
            "-" * 15,
        )  # separating line
        table += "{:<15} | {:<15}\n".format("Name", self.summary.name)
        table += "{:<15} | {:<15}\n".format("Errors", str(self.summary.errors))
        table += "{:<15} | {:<15}\n".format(
            "Failures",
            str(self.summary.failures),
        )
        table += "{:<15} | {:<15}\n".format(
            "Skipped",
            str(self.summary.skipped),
        )
        table += "{:<15} | {:<15}\n".format("Tests", str(self.summary.tests))
        table += "{:<15} | {:<15}\n".format("Time", str(self.summary.time))
        table += "{:<15} | {:<15}\n".format(
            "Timestamp",
            str(self.summary.timestamp),
        )
        table += "{:<15} | {:<15}\n".format("Hostname", self.summary.hostname)

        return table


@dataclass(frozen=True)
class SolutionMetric:
    is_success: bool | None
    """Whether this result is considered a success. If there is no definition
    of success (for example a certain edit-distance or CodeBLEU), then this
    might be None"""
    float_val: float
    """A float representation of the metric"""
    label: str
    """The kind of this metric"""


@dataclass(frozen=True)
class SolutionEvaluation:
    solution: CodeSolution
    test_results: TestRunResultList | None
    extra_metrics: list[SolutionMetric]
    main_metric: SolutionMetric | None
    exception: Exception | None = None
    exception_traceback: str | None = None

    @property
    def has_exception(self) -> bool:
        """Whether this evaluation did not complete do to an unexpected exception"""
        return self.exception is not None

    def get_metric_from_key(self, key: str) -> SolutionMetric | None:
        if self.main_metric and self.main_metric.label == key:
            return self.main_metric
        for metric in self.extra_metrics:
            if metric.label == key:
                return metric
        return None


MetricFunc = Callable[[CodeSolution], SolutionMetric | Sequence[SolutionMetric]]


def exact_match_metric(solution: CodeSolution, strip: bool = False) -> SolutionMetric:
    solve_marks = {
        (step.path, step.mark_id): step.value for step in solution.solve_steps
    }

    known_solution_marks = {
        (step.path, step.mark_id): step.value
        for sol in solution.problem.known_solutions
        for step in sol.solve_steps
    }

    if strip:
        solve_marks = {key: value.strip() for key, value in solve_marks.items()}
        known_solution_marks = {
            key: value.strip() for key, value in known_solution_marks.items()
        }

    exact_match = len(solve_marks) == len(known_solution_marks) and all(
        solve_marks.get(key) == known_solution_marks.get(key)
        for key in solve_marks.keys()
    )

    return SolutionMetric(
        is_success=exact_match,
        float_val=1.0 if exact_match else 0.0,
        label="exact_match",
    )


def create_exact_match_metric(strip=True) -> Callable:
    return lambda solution: exact_match_metric(solution, strip=strip)


class SolutionEvaluator:
    """
    Used to evaluate a solution made by a solver. It is configurable (and inherritable)
    in case want to try other metrics.
    """

    def __init__(
        self,
        metric_funcs: tuple[MetricFunc, ...] | None = None,
        do_not_execute_if_syntax_fail: bool = True,
        use_execution_cache: bool = True,
    ):
        self._extra_metric_funcs = metric_funcs
        self._do_not_execute_if_syntax_fail = do_not_execute_if_syntax_fail
        self._use_execution_cache = use_execution_cache

    @classmethod
    def default_instance(cls):
        return _solution_evaluator_default

    def evaluate(self, solution: CodeSolution) -> SolutionEvaluation:
        if solution.problem.test_cases:
            test_results = evaluate_code_problem_execution(
                solution.problem,
                solution,
                do_not_execute_if_syntax_fail=self._do_not_execute_if_syntax_fail,
                try_cache=self._use_execution_cache,
            )
        else:
            test_results = None
        spec = solution.problem.transformation_spec
        if len(list(spec.classification_question_statements())) > 0:
            classification_metrics = classification_metric_function(solution)
        else:
            classification_metrics = []
        extra_metrics = [*classification_metrics]
        if self._extra_metric_funcs:
            for metric_func in self._extra_metric_funcs:
                metric = metric_func(solution)
                if isinstance(metric, SolutionMetric):
                    extra_metrics.append(metric)
                else:
                    extra_metrics.extend(metric)
        return SolutionEvaluation(
            solution,
            test_results,
            extra_metrics,
            test_results.as_metric() if test_results is not None else extra_metrics[0],
        )


_solution_evaluator_default = SolutionEvaluator(None)


def classification_metric_function(
    solution: CodeSolution,
) -> Sequence[SolutionMetric]:
    classifications = list(
        solution.problem.transformation_spec.classification_question_statements(),
    )
    if len(classifications) == 0:
        return []
    if len(classifications) > 1:
        msg = "Multiple classification questions not supported"
        raise NotImplementedError(msg)
    classification = classifications[0].verb
    assert isinstance(classification, StsCodeClassificationQuestion)
    if not classification.operation_is_static():
        msg = "Classification question must be static"
        raise ValueError(msg)

    chosen_answer = get_only_item(
        [
            solve_step
            for solve_step in solution.solve_steps
            if solve_step.mark_id == classification.static_mark_id
        ],
    ).value
    chosen_answer = classification.value_to_prob_dict(chosen_answer)

    if solution.problem.known_solutions is None:
        return []
    solution_answer = get_only_item(
        [
            solve_step
            for solve_step in solution.problem.known_solutions[0].solve_steps
            if solve_step.mark_id == classification.static_mark_id
        ],
    ).value
    solution_answer = classification.value_to_prob_dict(solution_answer)

    # find the key with the highest value
    chosen_answer_key = max(chosen_answer, key=chosen_answer.get)
    solution_answer_key = max(solution_answer, key=solution_answer.get)
    match_correct = chosen_answer_key == solution_answer_key

    # Correct probability
    correct_answer_prob = chosen_answer[solution_answer_key]
    cross_entropy = -sum(
        [
            solution_answer[key] * math.log(chosen_answer[key] + 1e-10)
            for key in solution_answer
        ],
    )

    return [
        SolutionMetric(
            is_success=match_correct,
            float_val=correct_answer_prob,
            label="classification_correct_prob",
        ),
        SolutionMetric(
            is_success=match_correct,
            float_val=1.0 if match_correct else 0.0,
            label="classification_correct",
        ),
        SolutionMetric(
            is_success=match_correct,
            float_val=cross_entropy,
            label="classification_cross_entropy",
        ),
    ]


def evaluate_code_problem_execution(
    code_problem: CodeProblem,
    solution: CodeSolution,
    user_prompted_guard: bool = False,
    do_not_execute_if_syntax_fail: bool = True,
    try_cache: bool = False,
) -> TestRunResultList:
    """
    Takes a code problem and formats it with a solution. Then it takes all
    the test cases and assembles them into pytest code to run.
    """
    start = timer()
    syntax_check = is_solution_syntax_likely_valid(solution)
    if (not syntax_check) and do_not_execute_if_syntax_fail:
        return TestRunResultList(code_problem, syntax_check_result=syntax_check)
    if any(isinstance(tc, DiscoveredTestsuite) for tc in code_problem.test_cases):
        if len(code_problem.test_cases) != 1:
            msg = (
                "Currently does not support mixing discovered test cases with"
                " other test cases"
            )
            raise NotImplementedError(
                msg,
            )
        tc = code_problem.test_cases[0]

        if not isinstance(tc, DiscoveredTestsuite):
            msg = "Test case is not of DiscoveredTestsuite type"
            raise TypeError(msg)

        assert isinstance(tc, DiscoveredTestsuite)
        exec_result = _evaluate_suite_problem(
            solution,
            tc,
            user_prompted_guard,
            try_cache=try_cache,
        )
    else:
        exec_result = _evaluate_self_contained_problem(
            solution,
            code_problem.test_cases,
            user_prompted_guard,
            try_cache=try_cache,
        )
    result = parse_junit_test_cases(
        exec_result.xml_result, code_problem)
    result.syntax_check_result = syntax_check
    result.runtime = timer() - start
    result.collection_error = exec_result.collection_error
    result.exec_error = exec_result.exec_error
    if exec_result.test_suite_exec_result is not None:
        result.timeout = exec_result.test_suite_exec_result.timeout
    return result


def is_solution_syntax_likely_valid(solution: CodeSolution) -> bool:
    """Hacky just check of the syntax of a module"""
    new_files = apply_solution(solution)
    dirty_files = new_files.walk(include_dirs=False, only_consider_dirty=True)
    is_valid = True
    for path, file in dirty_files:
        lang_spec = solution.problem.get_lang_spec_for_path(path)
        is_valid &= lang_spec.check_no_syntax_errors(file.content_str)
    return is_valid


def _evaluate_suite_problem(
    solution: CodeSolution,
    discovered_suite: DiscoveredTestsuite,
    user_prompted_guard: bool,
    try_cache: bool = False,
) -> TestsuiteExecResult:
    if user_prompted_guard:
        msg = "User prompted guard not supported for suite problems"
        raise ValueError(msg)
    new_files = apply_solution(solution)

    problem = solution.problem
    docker_env = problem.environment.docker_env or PY_DEFAULT_DOCKER_ENV
    limits = docker_env.default_limits or ExecLimits()
    if try_cache:
        cache_key = (
            f"{new_files.get_hash()}{limits!s}{docker_env.get_hash()}"
            f"{discovered_suite.get_hash()}"
        )
        cached_result = eval_cache.get(cache_key)
        if cached_result:
            #print("Cache hit")
            #assert isinstance(cached_result, TestsuiteExecResult)
            #print(cached_result.test_suite_exec_result.cmd.cmd)
            #print(cached_result.test_suite_exec_result.stderr.decode()[:5000])
            return cached_result

    cmd_outputs = run_on_docker(
        docker_context=docker_env,
        cmds=discovered_suite.cmds,
        files=new_files,
        limits=limits,
        interactive_shell_debug=False,
    )
    # for cmd_output in cmd_outputs:
    #     import colorama
    #     # print the command in green
    #     print(colorama.Fore.GREEN + "$ " + cmd_output.cmd.cmd)
    #     # print the stdout in white
    #     print(colorama.Fore.WHITE + cmd_output.stdout.decode())
    #     # print the stderr in red
    #     print(colorama.Fore.RED + cmd_output.stderr.decode())
    #     # reset the colors back to normal
    #     print(colorama.Style.RESET_ALL)
    result = discovered_suite.cmds_results_to_suite(cmd_outputs)

    if try_cache:
        eval_cache[cache_key] = result

    return result


def _evaluate_self_contained_problem(
    solution: CodeSolution,
    test_cases: list[TestCase],
    user_prompted_guard: bool,
    try_cache: bool = False,
) -> TestsuiteExecResult:
    new_files = apply_solution(solution)
    code_problem = solution.problem
    step = solution.solve_steps[0]
    edit_path = step.path
    all_tests_src = [
        test_case.format_for_test_framework(lang_spec=PythonLangSpec())
        for test_case in test_cases
    ]
    all_tests_src = "\n\n".join(all_tests_src)
    runable_code = new_files[edit_path].content_str  # TODO: avoid redecoding somehow
    runnable_code_with_tests = runable_code + "\n\nimport pytest\n" + all_tests_src
    if user_prompted_guard:
        pretty_print_python_code(runnable_code_with_tests)
        if input("Run? [y/N] ").lower() != "y":
            msg = "User aborted"
            raise ValueError(msg)
    new_files = new_files.set_file_contents(edit_path, runnable_code_with_tests)
    use_docker_env = code_problem.environment.docker_env or PY_DEFAULT_DOCKER_ENV
    if try_cache:
        cache_key = f"{new_files.get_hash()}{edit_path}{use_docker_env.get_hash()}"
        cached_result = eval_cache.get(cache_key)
        if cached_result:
            return cached_result

    result = pytest_on_docker(
        files=new_files,
        run_file=edit_path,
        docker_env=use_docker_env,
    )
    if try_cache:
        eval_cache[cache_key] = result
    return result


def _group_steps_by_path(
    steps: Sequence[SolveStep],
) -> dict[str, list[SolveStep]]:
    path_to_steps = defaultdict(list)
    for step in steps:
        path_to_steps[step.path].append(step)
    return path_to_steps


def apply_steps_to_markup(
    content_node: StsMarkNode,
    steps: Sequence[SolveStep],
    still_hide_invisible: bool = False,
    stop_after_first_change: bool = False,
) -> str:
    key_to_step = {step.mark_id: step for step in steps}
    if len(key_to_step) != len(steps):
        msg = "The same mark_id was used twice in a solution"
        raise ValueError(msg)
    text = []
    visitor = content_node.iterate_parser_states()
    try:
        state = next(visitor)
    except StopIteration:
        state = None
    while state:
        if state is None:
            break
        visit_children = True
        node = state.node
        if isinstance(node, MarkElement) and node.mark_id in key_to_step:
            value = key_to_step[node.mark_id].value
            _verify_solution_value_is_valid_place_transforming(node, value)
            text.append(value)
            if stop_after_first_change:
                break
            visit_children = False
        elif isinstance(node, MarkText):
            if not still_hide_invisible or state.is_visible:
                text.append(node.text)
        try:
            state = visitor.send(visit_children)
        except StopIteration:
            state = None
    return "".join(text)


def make_mark_id_to_node(
    root_node: StsMarkNode,
) -> dict[str, MarkElement]:
    visitor = root_node.depth_first_iterate()
    mark_id_to_node = {}
    try:
        node = next(visitor)
    except StopIteration:
        node = None
    while node:
        if isinstance(node, MarkElement):
            mark_id_to_node[node.mark_id] = node
        try:
            node = visitor.send(True)
        except StopIteration:
            node = None
    return mark_id_to_node


def _verify_solution_value_is_valid_place_transforming(node, value):
    if node.verb.max_length_chars is not None:
        raise NotImplementedError
    _verify_max_length_lines(node, value)

    if not isinstance(node.verb, StsPlaceTransforming):
        msg = "Node verb is not of StsPlaceTransforming type"
        raise TypeError(msg)

    _verify_end_text(node, value)
    _verify_space_indent(node, value)
    # TODO: verify respect the the stop_at_block_end flag


def _verify_max_length_lines(node, value):
    if (
        node.verb.max_length_lines is not None
        and len(value.splitlines()) > node.verb.max_length_lines
    ):
        msg = f"Too many lines in the solution for {node.mark_id}"
        raise ValueError(msg)


def _verify_space_indent(node, value):
    verb = node.verb
    if verb.set_base_space_indent:
        first_line = get_first_line_with_nonwhitespace(value)
        if first_line.strip() != "" and not first_line.startswith(
            " " * verb.set_base_space_indent
        ):
            msg = (
                f"Solution for {node.mark_id} must start with"
                f" {verb.set_base_space_indent} spaces.\n"
                " Value:\n"
                f" {value!r}"
            )
            raise ValueError(
                msg,
            )


def _verify_end_text(node, value):
    if node.verb.set_ending_text and not value.endswith(
        node.verb.set_ending_text,
    ):
        escaped_expected = node.verb.set_ending_text.replace("\n", "\\n")
        msg = f"Solution for {node.mark_id} must end with '{escaped_expected}'"
        raise ValueError(
            msg,
        )


def apply_solution(
    solution: CodeSolution,
    still_hide_invisible: bool = False,
    stop_after_first_change: bool = False,
) -> ProjectDir:
    """Produces new working directory files after apply the steps in a solution"""
    problem = solution.problem
    path_to_steps = _group_steps_by_path(solution.solve_steps)
    wd = problem.working_directory
    files = wd.files
    for path, steps in path_to_steps.items():
        markup_node = markup_path(wd, path, problem.transformation_spec)
        new_content = apply_steps_to_markup(
            markup_node,
            steps,
            still_hide_invisible=still_hide_invisible,
            stop_after_first_change=stop_after_first_change,
        )
        files = files.set_file_contents(path, new_content)
    return files


junit_xml_schema = (Path(__file__) / "../JUnit.xsd").resolve().read_bytes()


def parse_junit_test_cases(
    xml_text: str,
    problem: CodeProblem,
) -> TestRunResultList:
    results = TestRunResultList(problem)
    if xml_text == "":
        logger.debug("EMPTY XML!!!!")
        return results
    schema_root = etree.XML(junit_xml_schema)
    schema = etree.XMLSchema(schema_root)

    strict_parser = etree.XMLParser(schema=schema, dtd_validation=False)
    xml_valid = False
    # ValueError: Unicode strings with encoding declaration are not supported.
    # Please use bytes input or XML fragments without declaration.
    xml_bytes = xml_text.encode(encoding="utf-8")
    try:
        etree.fromstring(xml_bytes, strict_parser)
        xml_valid = True
    except etree.XMLSyntaxError:
        xml_valid = False
        logger.warning("Error! JUnit test result is not valid XML")

    junit_xml = etree.fromstring(xml_bytes)
    testsuite = junit_xml.find("testsuite")
    if testsuite is None:
        # Deal with the java junit output where the root is just the testsuite
        #   and no namesapce for the xml
        testsuite = junit_xml.xpath("//*[local-name() = 'testsuite']")
        assert len(testsuite) == 1
        testsuite = testsuite[0]

    if testsuite.get("timestamp", None) is None:
        timestamp = None
    else:
        timestamp = datetime.fromisoformat(testsuite.get("timestamp"))
    testsuite_summary = TestRunSummary(
        name=testsuite.get("name"),
        errors=int(testsuite.get("errors", 0)),
        failures=int(testsuite.get("failures", 0)),
        skipped=int(testsuite.get("skipped", 0)),
        tests=int(testsuite.get("tests")),
        time=timedelta(seconds=float(testsuite.get("time", 0))),
        timestamp=timestamp,
        hostname=testsuite.get("hostname", ""),
        valid_xml=xml_valid,
    )
    results.summary = testsuite_summary

    for case in junit_xml.iter("testcase"):
        fail_val = case.find("failure")
        if fail_val is None:
            fail_val = case.find("error")
        passed = fail_val is None
        results.append(
            TestRunResult(
                success=passed,
                output=None,
                completed=True,
                fail_message=(
                    fail_val.attrib.get("message", None)
                    if fail_val is not None
                    else None
                ),
                fail_type=fail_val.get("type", None) if fail_val is not None else None,
                fail_text=fail_val.text if fail_val is not None else None,
                id_=case.get("id", None),
                name=case.get("name", None),
            ),
        )
    return results


def pull_value_answers_from_solution(
    solution: CodeSolution,
) -> list[tuple[MarkElement[StsValueQuery], str]]:
    problem = solution.problem
    path_to_steps = _group_steps_by_path(solution.solve_steps)
    wd = problem.working_directory
    values = []
    for path, steps in path_to_steps.items():
        question_steps = [
            step for step in steps if isinstance(step.value, StsValueQuery)
        ]
        markup_node_root = markup_path(wd, path, problem.transformation_spec)
        mark_id_to_node = make_mark_id_to_node(markup_node_root)
        for step in question_steps:
            node = mark_id_to_node[step.mark_id]
            values.append((node, step.value))
    return values



import argparse
import json
import math
import os
import re
from pathlib import Path
from time import sleep

import tiktoken
from lmwrapper.openai_wrapper import OpenAiModelNames, OpenAIPredictor, get_open_ai_lm
from lmwrapper.structs import LmPrompt

# by default should be defects4j. In some cases, it might be <PATH_TO_DEFECTS4J>/framework/bin/defects4j
DEFECTS4J_COMMAND = "defects4j"


def preprocess(data):
    # creates the Defects4J_projects directory and checks out the buggy and fixed versions of the project

    print("Checking out buggy and fixed versions of projects...")
    # check if the Defects4J_projects directory exists
    if not Path("Defects4J_projects").exists():
        # if it doesn't exist, create it
        Path("Defects4J_projects").mkdir()

    for project in data:
        print(project)
        # this project is not deprecated
        # check if the buggy version of the project exists
        if not Path("Defects4J_projects/" + project[0] + "_" + project[1]).exists():
            # if it doesn't exist, check it out
            os.system(
                DEFECTS4J_COMMAND
                + " checkout -p "
                + project[0]
                + " -v "
                + project[1]
                + "b -w Defects4J_projects/"
                + project[0]
                + "_"
                + project[1]
                + "_buggy",
            )

        # check if the fixed version of the project exists
        if not Path(
            "Defects4J_projects/" + project[0] + "_" + project[1] + "_fixed",
        ).exists():
            # if it doesn't exist, check it out
            os.system(
                DEFECTS4J_COMMAND
                + " checkout -p "
                + project[0]
                + " -v "
                + project[1]
                + "f -w Defects4J_projects/"
                + project[0]
                + "_"
                + project[1]
                + "_fixed",
            )

        # copy over .defects4j.config from the fixed version to the buggy version
        # this is done because of an edge case where the fixed version removes some tests that are present in the buggy version
        # this causes defects4j to fail when running the test suite even though the buggy version should be passing
        os.system(
            "cp Defects4J_projects/"
            + project[0]
            + "_"
            + project[1]
            + "_fixed/.defects4j.config Defects4J_projects/"
            + project[0]
            + "_"
            + project[1]
            + "_buggy/.defects4j.config",
        )


def test_passing_data(project):
    # save the current working directory
    cwd = os.getcwd()

    # navigate to the buggy version of the project
    os.chdir("Defects4J_projects/" + project[0] + "_" + project[1] + "_buggy")

    try:
        # this block indicates a successful code change
        # run the test suite and pipe the output to a file
        os.system(
            DEFECTS4J_COMMAND
            + " test > "
            + project[0]
            + "_"
            + project[1]
            + "_buggy_test_output.txt",
        )

        # open the file
        with open(
            project[0] + "_" + project[1] + "_buggy_test_output.txt",
            encoding="locale",
        ) as file:
            # extract the number of failing tests. The number of failing tests is after the phrase "Failing tests: "
            failing_tests = file.read().split("Failing tests: ")[1].split("\n")[0]
            print("Failing tests: " + failing_tests)

        # get the total number of tests
        # this is the length of the file "all_tests"
        with open("all_tests", encoding="locale") as file:
            total_tests = len(file.readlines())
            print("Total tests: " + str(total_tests))

        passing_tests = total_tests - int(failing_tests)

    except:
        # this block indicates a compilation error
        # we go into this block if the test suite fails to run

        # navigate into the fixed version of the project
        os.chdir("../" + project[0] + "_" + project[1] + "_fixed")
        os.system(
            DEFECTS4J_COMMAND
            + " test > "
            + project[0]
            + "_"
            + project[1]
            + "_fixed_test_output.txt",
        )

        passing_tests = 0

        with open("all_tests", encoding="locale") as file:
            total_tests = str(len(file.readlines()))
            print("Total tests: " + total_tests)

    # navigate back to the original directory
    os.chdir(cwd)

    return passing_tests, total_tests


def check_equal_sign(line):
    return (
        "=" in line
        and "==" not in line
        and "!=" not in line
        and "<=" not in line
        and ">=" not in line
        and "=>" not in line
    )


def main():
    parser = argparse.ArgumentParser()
    # Required parameters
    parser.add_argument(
        "--model",
        default=None,
        type=str,
        required=True,
        help="codex/turbo",
    )

    args = parser.parse_args()

    if args.model == "codex":
        lm = get_open_ai_lm(
            OpenAiModelNames.code_davinci_002,
        )
        OpenAIPredictor.configure_global_ratelimit(20, per_seconds=60)
    if args.model == "turbo":
        lm = get_open_ai_lm(
            OpenAiModelNames.gpt_3_5_turbo_instruct,
        )

    import csv

    # Set a larger field size limit, e.g., 2 MB (you can adjust this value as needed)
    csv.field_size_limit(10 * 1024 * 1024)

    # Specify the file path
    file_path = "Defects4J_oneLiner_metadata.csv"

    # Initialize an empty list to store the data
    data = []

    # Open the CSV file and read its contents
    with open(file_path, newline="", encoding="locale") as csvfile:
        csvreader = csv.reader(csvfile)

        # Loop through each row in the CSV file
        for row in csvreader:
            data.append(row)

    preprocess(data)

    ff = open(args.model + "_result_log_prob.txt", "a", encoding="locale")
    ff.write(
        "buggy"
        + "\t\t"
        + "Fixed"
        + "\t\t"
        + "generated_text"
        + "\t\t"
        + "is_exact"
        + "\t\t"
        + "total_prob"
        + "\t\t"
        + "average_prob"
        + "\t\t"
        + "log_probs"
        + "\t\t"
        + "length"
        + "\t\t"
        + "total_tests"
        + "\t\t"
        + "passing_tests"
        + "\n",
    )
    ff.close()

    enc = tiktoken.encoding_for_model("code-davinci-002")

    error_count = 0
    i = 0
    correct = 0
    json_objects = []
    while i < len(data):
        try:
            print(i)
            print(data[i])
            is_exact = 0
            file_name = data[i][2]
            line_number = int(data[i][3])

            # create a Path of Defects4J_projects/data[i][0]_data[i][1]_buggy
            buggy_file = (
                Path("Defects4J_projects/" + data[i][0] + "_" + data[i][1] + "_buggy")
                / file_name
            )
            fixed_file = (
                Path("Defects4J_projects/" + data[i][0] + "_" + data[i][1] + "_fixed")
                / file_name
            )

            print("Buggy file: ", buggy_file)
            print("Fixed file: ", fixed_file)

            # navigate to the buggy file and print out the contents of line at line_number
            with open(buggy_file, encoding="locale") as file:
                buggy_file_contents = file.readlines()
                buggy_line = buggy_file_contents[line_number - 1]
                print("Buggy line: ", buggy_line)

                # create a prompt for the buggy line
                # #############################
                before_buggy_line = buggy_file_contents[:line_number]
                after_buggy_line = buggy_file_contents[line_number:]
                # take last element of list split it by "=" and add <buggy> tag to the last element
                # only if the line contains "=" but not "==", "!=", "<=", ">=" and not "=>"
                if check_equal_sign(before_buggy_line[-1]):
                    before_buggy_line[-1] = (
                        before_buggy_line[-1].strip().split("=")[0]
                        + "= "
                        + "<buggy>"
                        + before_buggy_line[-1].strip().split("=")[1]
                        + "</buggy>"
                    )
                else:
                    before_buggy_line[-1] = (
                        "<buggy>" + before_buggy_line[-1] + "</buggy>"
                    )
                # convert list to string
                before_buggy_line = "".join(before_buggy_line)
                before_buggy_line = enc.encode(before_buggy_line)
                before_buggy_line = before_buggy_line[
                    len(before_buggy_line) - 200 : len(before_buggy_line)
                ]
                before_buggy_line = enc.decode(before_buggy_line)
                # convert list to string
                after_buggy_line = "".join(after_buggy_line)
                after_buggy_line = enc.encode(after_buggy_line)
                after_buggy_line = after_buggy_line[0:200]
                after_buggy_line = enc.decode(after_buggy_line)
                # join the two strings
                prompt = before_buggy_line + after_buggy_line + "\n"
                json_objects.append(
                    {
                        "project": data[i][0],
                        "number": data[i][1],
                        "prompt": prompt,
                    },
                )
                prompt += (
                    "Question: There is a bug in the above code snippet tagged by"
                    " <buggy> and </buggy>. Please generate the correct version.\n"
                )
                before_fixed_line = before_buggy_line.split("<buggy>")[0] + "<fixed>"
                prompt += "Answer:\n" + before_fixed_line
                # #############################

            # navigate to the fixed file and print out the contents of line at line_number
            with open(fixed_file, encoding="locale") as file:
                fixed_file_contents = file.readlines()
                fixed_line = fixed_file_contents[line_number - 1]

            print("Buggy line: ", buggy_line)
            print("Fixed line: ", fixed_line)
            print("Prompt: ", prompt)

            pred = lm.predict(
                LmPrompt(
                    text=prompt,
                    max_tokens=200,
                    logprobs=5,
                    stop=["\n", "</fixed>"],
                    num_completions=1,
                    temperature=0,
                ),
            )

            text = pred.completion_text.strip()
            log_probs = pred.completion_logprobs

            print(text)
            print(log_probs)

            probs = [math.exp(log_prob) for log_prob in log_probs]
            print(probs)
            average = sum(probs) / len(log_probs)
            print(average)
            total = math.exp(sum(log_probs))
            print(total)

            print("Total_Probability: ", total)
            print("Average_Probability: ", average)
            print("Log_Probabilities: ", log_probs)
            print("Length: ", len(text))

            # reopen the buggy file and replace the buggy line with the fixed line
            with open(buggy_file, "r+", encoding="locale") as file:
                print("Replacing buggy line with fixed line...")
                buggy_file_contents = file.readlines()
                if check_equal_sign(buggy_file_contents[line_number - 1]):
                    buggy_file_contents[line_number - 1] = (
                        buggy_file_contents[line_number - 1].strip().split("=")[0]
                        + "= "
                        + text
                        + "\n"
                    )
                    if re.sub(
                        r"\s",
                        "",
                        buggy_file_contents[line_number - 1],
                    ) == re.sub(r"\s", "", fixed_line):
                        is_exact = 1
                        correct = correct + 1
                else:
                    buggy_file_contents[line_number - 1] = text + "\n"
                    if re.sub(
                        r"\s",
                        "",
                        buggy_file_contents[line_number - 1],
                    ) == re.sub(r"\s", "", fixed_line):
                        is_exact = 1
                        correct = correct + 1
                print(buggy_file_contents[line_number - 1])

                with open(buggy_file, "w", encoding="locale") as file:
                    print("Writing to file...")
                    file.writelines(buggy_file_contents)

            # get passing tests and total tests
            passing_tests, total_tests = test_passing_data(data[i])

            print("########")

            error_count = 0
            i = i + 1

            ff = open(args.model + "_result_log_prob.txt", "a", encoding="locale")
            ff.write(
                buggy_line.strip()
                + "\t\t"
                + fixed_line.strip()
                + "\t\t"
                + text.strip()
                + "\t\t"
                + str(is_exact)
                + "\t\t"
                + str(total)
                + "\t\t"
                + str(average)
                + "\t\t"
                + str(log_probs)
                + "\t\t"
                + str(len(text))
                + "\t\t"
                + str(total_tests)
                + "\t\t"
                + str(passing_tests)
                + "\n",
            )
            ff.close()

        except Exception as e:
            print(e)
            error_count = error_count = error_count + 1
            if error_count == 12:
                i = i + 1
                error_count = 0
                ff = open(args.model + "_result_log_prob.txt", "a", encoding="locale")
                ff.write("\n")
                ff.close()

            sleep(5)

            continue

    # to dump the prompts used to generate the results
    with open(args.model + "_result_log_prob.jsonl", "w", encoding="locale") as f:
        json.dump(json_objects, f, indent=4)


main()



"""
A PromptRenderer and parser that prompts a model with xml-like
tags on places to insert text. The model is then expected to output
the corresponding edit tags and locations. We expect this to
pretty hard.
"""

import dataclasses
import warnings
from io import StringIO

from lmwrapper.abstract_predictor import LmPredictor
from lmwrapper.structs import LmPrediction, LmPrompt
from lxml import etree

from synthegrator.code_problems import CodeProblem
from synthegrator.environments import ProjectWorkingDirectory
from synthegrator.few_shotting import FewShotConfig
from synthegrator.problem_rendering import LmPromptRender, PromptRenderer
from synthegrator.prompting_test_case_selection import (
    PromptingTestCaseSelectionStrategy,
)
from synthegrator.response_parser import (
    ResponseParser,
    format_return_val_for_node,
)
from synthegrator.transformation_spec import (
    MarkElement,
    MarkText,
    SolveStep,
    StsEditable,
    StsInsert,
    StsPlaceTransforming,
    TransformationSpec,
    get_mark_element,
    get_verbs_per_path,
    map_paths_to_path_ids,
    markup_path,
)


class LmPromptRenderMultiEdit(LmPromptRender):
    pass


@dataclasses.dataclass(frozen=True)
class LmTaggedEditPrompt(LmPromptRenderMultiEdit):
    tag_name_edit: str
    tag_name_solve: str
    preprompted_tag_start: str = ""


class PromptRendererMultiEdit(PromptRenderer):
    def __init__(self):
        super().__init__()

    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> "LmPromptRenderMultiEdit":
        raise NotImplementedError

    def __call__(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> "LmPromptRenderMultiEdit":
        return self.render(problem, lm, prompt_seed)


class TaggedEditRenderer(PromptRendererMultiEdit):
    """Renders a problem as a sequence of edits in xml-like tags"""

    def __init__(
        self,
        few_shot_config: FewShotConfig = None,
        prompt_test_case_selection_strategy: PromptingTestCaseSelectionStrategy = None,
        tag_name_edit: str = "edit",
        tag_name_solve: str = "edit_solve",
        include_first_tag_at_end: bool | None = None,
        custom_closing_lines: str | None = None,
        add_stop: bool | None = None,
    ):
        super().__init__()
        self._few_shot_config = few_shot_config
        self._prompt_test_case_selection_strategy = prompt_test_case_selection_strategy
        self._tag_name_edit = tag_name_edit
        self._tag_name_solve = tag_name_solve
        self._include_first_tag_at_end = include_first_tag_at_end
        self._custom_closing_lines = custom_closing_lines
        self._add_stop = add_stop

    def render(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> LmTaggedEditPrompt:
        if self._few_shot_config is not None:
            warnings.warn(
                "Few shot not supported with this non-chat version. "
                "No few shot will be used. TODO",
            )
        spec = problem.transformation_spec
        if spec.count_editing_statements() != 1:
            msg = "Currently simple and assumes a exactly one edit statement"
            raise NotImplementedError(
                msg,
            )
        path_to_verbs = get_verbs_per_path(problem.working_directory, spec)
        prompt_lines = []
        if problem.instructions:
            prompt_lines.append(problem.instructions)
        path_to_path_id = map_paths_to_path_ids(problem.working_directory, spec)
        all_tag_ids = []
        for path in path_to_verbs:
            markedup_text = markup_path(
                problem.working_directory,
                path,
                problem.transformation_spec,
            )

            prompt_lines.append("@@ " + str(path) + " @@")
            tagged_prompt, tag_ids = _marked_up_text_to_tagged_prompt_text(
                markedup_text,
                path_id=path_to_path_id[path],
                tag_name=self._tag_name_edit,
            )
            prompt_lines.append(tagged_prompt)
            all_tag_ids.extend(tag_ids)
            prompt_lines.append("")
        prompt_lines.append("---")
        closing, prepend = _make_closing_lines(
            all_tag_ids=all_tag_ids,
            tag_name_edit=self._tag_name_edit,
            tag_name_solve=self._tag_name_solve,
            include_first_tag_at_end=(
                not lm.is_chat_model
                if self._include_first_tag_at_end is None
                else self._include_first_tag_at_end
            ),
            custom_closing_lines=self._custom_closing_lines,
        )
        prompt_lines.append(closing)
        text = "\n".join(prompt_lines)
        stop = None
        add_stop = self._add_stop
        if add_stop is None:
            add_stop = not lm.is_chat_model
        if len(all_tag_ids) == 1 and add_stop:
            stop = [f"</{self._tag_name_solve}>"]
        return LmTaggedEditPrompt(
            prompt=LmPrompt(
                text=text,
                stop=stop,
            ),
            tag_name_edit=self._tag_name_edit,
            tag_name_solve=self._tag_name_solve,
            preprompted_tag_start=prepend,
        )

    def __call__(
        self,
        problem: CodeProblem,
        lm: LmPredictor,
        prompt_seed: int | None = None,
    ) -> "LmTaggedEditPrompt":
        return self.render(problem, lm, prompt_seed)


def _make_closing_lines(
    all_tag_ids: list[str],
    tag_name_edit: str = "edit",
    tag_name_solve: str = "edit_solve",
    include_first_tag_at_end: bool = True,
    custom_closing_lines: str | None = None,
) -> tuple[str, str]:
    value = (
        f"Implement the correct changes in the above code in the <{tag_name_edit}>"
        f' tags. Please place each fix in a separate <{tag_name_solve} id="id">'
        f' tags.\nFor example:\n<{tag_name_solve} id="id_example">new replacement for'
        f" the text...</{tag_name_solve}>\nAnswer:\n"
    )
    if custom_closing_lines:
        value = custom_closing_lines
    if include_first_tag_at_end:
        preprepended_prompt = f'<{tag_name_solve} id="{all_tag_ids[0]}">'
    else:
        preprepended_prompt = ""
    return value + preprepended_prompt, preprepended_prompt


def _marked_up_text_to_tagged_prompt_text(
    markedup_text: MarkElement,
    path_id: str,
    tag_name: str = "edit",
) -> tuple[str, list[str]]:
    """
    Converts a marked up text to a prompt
    Returns the prompt text and a list of the ids of the tags in the prompt
    """
    text = []
    tag_ids = []
    for state in markedup_text.iterate_parser_states():
        node = state.node
        if not state.is_visible:
            continue
        if isinstance(node, MarkText):
            text.append(node.text)
        if isinstance(node, MarkElement):
            verb = node.verb
            if isinstance(verb, StsEditable | StsInsert):
                tag_id = f"{path_id}_{node.mark_id}"
                tag_ids.append(tag_id)
                tag_props = [f'id="{tag_id}"']
                if verb.max_length_lines:
                    tag_props.append(f"max_length_lines={verb.max_length_lines}")
                tag = f"<{tag_name} {' '.join(tag_props)}>"
                text.append(tag)

                def close_tag():
                    text.append(f"</{tag_name}>")

                state.after_visit_children_callbacks.append(close_tag)
            else:
                pass
                # raise NotImplementedError("Unhandled {}".format(node.verb))
    return "".join(text), tag_ids


class TaggedEditResponseParser(ResponseParser):
    def __init__(self):
        super().__init__()

    def parse(
        self,
        render: LmTaggedEditPrompt,
        resp: LmPrediction,
        problem: CodeProblem,
    ) -> list[SolveStep]:
        ids_to_values = _find_ids_and_values_in_tagged_answer(
            render.preprompted_tag_start + resp.completion_text,
            tag_name_edit=render.tag_name_edit,
            tag_name_solve=render.tag_name_solve,
            default_id=_get_mark_default_id(problem),
        )
        spec = problem.transformation_spec
        wd = problem.working_directory
        # print(resp.completion_text)
        ids_to_path_and_mark_id = _ids_to_path_id_and_mark_id(ids_to_values, spec, wd)
        solve_steps = []
        for tag_id, (path, mark_id) in ids_to_path_and_mark_id.items():
            node = get_mark_element(spec, path, wd, mark_id)
            if node is None:
                continue  # We are permissive for responses with nonexisted nodes
            lang_spec = problem.get_lang_spec_for_path(path)
            solution_text = format_return_val_for_node(
                ids_to_values[tag_id],
                node,
                lang_spec,
            )
            solve_steps.append(
                SolveStep(
                    path,
                    mark_id,
                    solution_text,
                ),
            )
        return solve_steps


def _find_ids_and_values_in_tagged_answer(
    answer: str,
    tag_name_edit: str = "edit",
    tag_name_solve: str = "edit_solve",
    default_id: str = None,
) -> dict[str, str]:
    """Returns a dictionary mapping from the id of each tag to its value"""
    # The lxml parser requires a root element to parse the content correctly.
    # We wrap the content in a root element named <root>.
    content_wrapped = f"<root>{answer}</root>"

    # Parse the content
    tree = etree.parse(StringIO(content_wrapped), etree.HTMLParser())

    # Initialize the dictionary to store ID to content mapping
    id_to_value = {}

    # Find all <edit_solve> elements and add them to the dictionary
    edit_solves = tree.xpath(f"//{tag_name_solve}")
    for edit_solve in edit_solves:
        edit_id = edit_solve.get("id", default_id)
        if edit_id:
            edit_content = (edit_solve.text or "")
            id_to_value[edit_id] = edit_content

    # We want to be extra permissive in case the model decides to use
    # the <edit> tag instead of the <edit_solve> tag.
    # Find all <edit> elements and add them to the dictionary
    # if their ID is not already present
    edits = tree.xpath(f"//{tag_name_edit}")
    for edit in edits:
        edit_id = edit.get("id", default_id if default_id not in id_to_value else None)
        if edit_id and edit_id not in id_to_value:
            edit_content = (edit.text or "")
            id_to_value[edit_id] = edit_content

    if len(id_to_value) == 0 and default_id is not None:
        striped = answer.split(f"</{tag_name_edit}>", 1)[0]
        striped = striped.split(f"</{tag_name_solve}>", 1)[0]
        id_to_value[default_id] = striped

    return id_to_value


def _ids_to_path_id_and_mark_id(
    ids_to_values: dict[str, str],
    spec: TransformationSpec,
    wd: ProjectWorkingDirectory,
) -> dict[str, tuple[str, str]]:
    path_ids_to_paths = {v: k for k, v in map_paths_to_path_ids(wd, spec).items()}
    ids_to_path_and_mark_id = {}
    for id in ids_to_values:
        if len(id.split("_")) != 2:
            continue
        path_id, mark_id = id.split("_")
        path = path_ids_to_paths.get(path_id)
        if path is None:
            continue
        ids_to_path_and_mark_id[id] = (path, mark_id)
    return ids_to_path_and_mark_id


def _get_mark_default_id(problem: CodeProblem) -> str | None:
    path_to_verbs = get_verbs_per_path(
        problem.working_directory, problem.transformation_spec
    )
    if len(path_to_verbs) > 1:
        raise NotImplementedError("Multiple paths not supported")
    for path, verbs in path_to_verbs.items():
        if not verbs:
            continue
        markup_text = markup_path(
            problem.working_directory, path, problem.transformation_spec
        )
        for state in markup_text.iterate_parser_states():
            if not state.is_visible:
                continue
            if isinstance(state.node, MarkElement):
                if isinstance(state.node.verb, StsPlaceTransforming):
                    return "0_" + state.node.mark_id
    return None



import dataclasses
import logging
logger = logging.getLogger(__name__)
import os
import random
from pathlib import Path
from random import randint

import pandas as pd
from tree_sitter import Node

from synthegrator.code_problem_builders import make_simple_line_edit_problem
from synthegrator.code_problems import (
    CodeProblem,
    DiscoveredTestsuite,
)
from synthegrator.lang_specs.lang_spec import (
    PythonFunctionParser,
)
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.sandboxing import DYPYBENCH_DOCKER_ENV, Cmd
from synthegrator.util import (
    IteratorWithLength,
    MultilineString,
)

cur_path = Path(__file__).parent.absolute()


def _default_projects_root() -> Path:
    ev = os.environ.get("DYPYBENCH_PROJECTS_ROOT", None)
    directory = None
    if ev is not None:
        directory = Path(ev).resolve()
    elif Path("~/dypybench_projects").expanduser().exists():
        directory = Path("~/dypybench_projects").expanduser().resolve()
    if directory is None or not directory.exists():
        if ev is not None:
            print(f"DYPYBENCH_PROJECTS_ROOT {ev} is not a valid directory")
        else:
            print(
                "DYPYBENCH_PROJECTS_ROOT is not set and ~/dypybench_projects does not exist. See "
                "synthegrator/synthdatasets/dypybench_setup.md for setup info on getting"
                "the projects to the source.",
            )
        msg = "Could not find the dypybench projects root"
        raise FileNotFoundError(msg)
    return directory


_suite_problems = [  # Precompute a test suite for each project.
    DiscoveredTestsuite(
        cmds=[
            # Cmd(
            #    "cd /DyPyBench && python3 dypybench.py --copy"
            # ),
            Cmd(
                "cd /DyPyBench && python3 dypybench.py --testtemp"
                f" {project_id} --target /sandbox",
            ),
            Cmd(f"cat /sandbox/project{project_id}/result.xml"),
        ],
    )
    for project_id in range(51)
]

_project_docker_envs = [
    dataclasses.replace(
        DYPYBENCH_DOCKER_ENV,
        before_file_copy_cmds=(
            *list(DYPYBENCH_DOCKER_ENV.before_file_copy_cmds),
            Cmd(
                f"cd /DyPyBench && python3 dypybench.py --copy {project_id} --target"
                " /sandbox",
            ),
        ),
    )
    for project_id in range(51)
]


def _sample_line(
    random_line_seed: int,
    path: Path,
    function: MultilineString,
    function_parser: PythonFunctionParser,
) -> int:
    sampled = 0
    path_seed = int.from_bytes(str(path).encode("utf-8"), "big") % 2**30
    while sampled <= 100:
        old_state = random.getstate()
        use_seed = random_line_seed + sampled + path_seed
        random.seed(use_seed)
        line_number = randint(
            function_parser.body_span_begin,
            function_parser.body_span_end,
        )
        random.setstate(old_state)
        line = function[line_number]
        if (
            "try:" not in line
            and line.strip() != ""
            and not line.lstrip().startswith("#")
        ):
            return line_number
        sampled += 1

    return None


def produce_dy_py_bench_problem(
    task_id: str,
    path: str,
    identifier: str,
    func_begin: int,
    func_end: int,
    function: str,
    project_id: str,
    line_to_replace: int | None = None,
    strip_docstring: bool = False,
    projects_root: str | None = None,
    random_line_seed: int = 42,
) -> CodeProblem:
    if projects_root is None:
        projects_root = _default_projects_root()

    projects_root = projects_root or _default_projects_root()
    func_begin = func_begin - 1  # 0 index
    func_end = func_end - 1  # 0 index

    function = MultilineString(function)
    function_parser = PythonFunctionParser(str(function))

    source_path = Path(projects_root) / f"project{project_id}" / path
    if not source_path.exists():
        msg = f"Could not find the source file for {source_path}"
        raise Exception(msg)

    source_file = MultilineString(source_path.read_text())

    if line_to_replace is None:
        function_line_to_replace = _sample_line(
            random_line_seed,
            source_path,
            function,
            function_parser,
        )
        if not function_line_to_replace:
            msg = f"Could not sample valid line for {task_id}"
            raise Exception(msg)
        source_line_to_replace = func_begin + function_line_to_replace
    else:
        source_line_to_replace = line_to_replace

    if strip_docstring:
        func_without_doc = MultilineString(
            function_parser.function_without_doc,
        )
        doc_offset = len(function) - len(func_without_doc)
        if line_to_replace:
            line_to_replace -= doc_offset
            assert line_to_replace > 0
            function_line_to_replace = line_to_replace
        function = func_without_doc

    known_solution = source_file[source_line_to_replace]

    invisible_spans = [
        (0, func_begin),
        (func_end + 1, len(source_file)),
    ]

    if strip_docstring:
        parser = function_parser
        docstring_node = parser.docstring_node
        if isinstance(docstring_node, Node):
            ds_start = func_begin + docstring_node.start_point[0]
            ds_end = func_begin + docstring_node.end_point[0] + 1
            if ds_end == invisible_spans[1][0]:
                logger.error("Is the function empty? %s", str(task_id))
                return None
            invisible_spans.insert(1, (ds_start, ds_end))
        else:
            logger.error("No docstring node? %s", str(task_id))

            return None

    return make_simple_line_edit_problem(
        code=str(source_file),
        invisible_spans=invisible_spans,
        test_cases=[_suite_problems[int(project_id)]],
        edit_lines_span=(source_line_to_replace, source_line_to_replace + 1),
        current_text_visible=False,
        target_path=f"project{project_id}/" + path,
        known_solutions=[known_solution],
        lang_spec=PythonLangSpec(),
        dataset_name="dypybench_line_completion",
        problem_id=task_id,
        docker_env=_project_docker_envs[int(project_id)],
        override_instructions="",
        cap_to_max_same_num_lines=True,
    )


def _gather_dypybench_funcs_df_from_json(
    json_path: str = cur_path / "dypybench_data/functions_dataset.jsonl",
    selected_ids_path: str = cur_path / "dypybench_data/selected_ids.parquet",
):
    df = pd.read_json(json_path, lines=True)
    if Path(selected_ids_path).exists():
        selected_ids = pd.read_parquet(selected_ids_path)
        df = df[df.index.isin(selected_ids.index)]
    df = df[
        (df.loc_len > 3)
        & (df.docstring_lines >= 1)
        & (df.coverage >= 100)
        & (df.filecoverage > 0)
    ]

    bad_functions = {
        109,
        7642,
        8226,
        8232,
        8239,
        8247,
        8255,
        8265,
        8306,
        8345,
        8384,
        8396,
        8457,
        8494,
        8507,
        8518,
        9068,
        9137,
        10776,
        10785,
        10789,
        10790,
        11438,
        20972,
        20973,
        20974,
        25916,
        26424,
        27592,
        28123,
    }

    # TODO: Remove bad functions from dataset itself rather than hack here
    return df[~df.index.isin(bad_functions)]


if __name__ == "__main__":
    df = _gather_dypybench_funcs_df_from_json()
    df.to_parquet(cur_path / "dypybench_data/func_dataset_subset.parquet")
    print(df)


def _get_dypy_df():
    return pd.read_parquet(cur_path / "dypybench_data/func_dataset_subset.parquet")


def yield_dypybench(
    df: pd.DataFrame = None,
    strip_docstring: bool = False,
    max_problems: int | None = None,
) -> IteratorWithLength[CodeProblem]:
    from datasets import Dataset

    if df is None:
        df = _get_dypy_df()

    dataset = Dataset.from_pandas(df, preserve_index=True).rename_column(
        "__index_level_0__",
        "index",
    )

    if max_problems and len(dataset) > max_problems:
        # Shuffle the dataset and take the first max_problems
        dataset = dataset.shuffle(seed=42).select(range(max_problems))

    def inner():
        for ex in dataset:
            problem = produce_dy_py_bench_problem(
                task_id=str(ex["index"]),
                path=str(ex["path"]),
                identifier=str(ex["identifier"]),
                func_begin=int(ex["func_begin"]),
                func_end=int(ex["func_end"]),
                function=str(ex["function"]),
                project_id=str(ex["project"]),
                line_to_replace=(
                    int(ex["line_number_to_replace"])
                    if ex.get("line_number_to_replace")
                    else None
                ),
                strip_docstring=strip_docstring,
            )

            if problem:
                yield problem

    return IteratorWithLength(inner(), len(dataset))



import dataclasses
from timeit import default_timer as timer

import numpy as np
from lmwrapper.abstract_predictor import LmPredictor
from lmwrapper.openai_wrapper import OpenAIPredictor
from lmwrapper.structs import (
    LmPrediction,
)

from synthegrator.code_problems import (
    CodeProblem,
    CodeSolution,
    LmCodeSolution,
)
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.problem_rendering import (
    LmPromptRenderSingleEdit,
    PromptRenderer,
    PromptRendererSingleEditGeneric,
)
from synthegrator.prompting_test_case_selection import (
    PromptingTestCaseSelectionStrategy,
)
from synthegrator.response_parser import ResponseParser, ResponseParserSingleEdit
from synthegrator.transformation_spec import (
    MarkElement,
    StsPlaceTransforming,
)
from synthegrator.uncertainty_modeling import (
    DelayedUncertaintyEstimate,
    PrecomputedUncertaintyEstimate,
    ProbabilityIsCorrect,
)
from synthegrator.util import (
    estimate_best_overlap_span,
)


class BaseCodeSolver:
    def solve(
        self,
        problem: CodeProblem,
        randomizing_seed: int | None = None,
    ) -> CodeSolution:
        raise NotImplementedError

    @property
    def allows_multithreading(self) -> bool:
        return False


class LmBasedSolver(BaseCodeSolver):
    """A solver that internally is based on a language model"""

    def __init__(self, model: LmPredictor, include_lm_response: bool = False):
        self._uncertainty_hooks = []
        self.model = model
        self.include_lm_response = include_lm_response

    def add_uncertainty_hook(self, hook: "LmUncertaintyHook"):
        self._uncertainty_hooks.append(hook)

    def _augment_solution_with_uncertainty_hooks(
        self,
        solution: CodeSolution,
        lm_prediction: LmPrediction,
        modified_full_text: str,
    ):
        solution.uncertainty_estimates.extend(
            [
                hook.prepare(
                    lm_prediction,
                    self.model,
                    solution,
                    modified_full_text,
                )
                for hook in self._uncertainty_hooks
            ],
        )

    @property
    def allows_multithreading(self) -> bool:
        if isinstance(self.model, OpenAIPredictor):
            return True
        return False


class DummyCodeSolverAutoRegressive(LmBasedSolver):
    """
    Solves a code problem by peaking at the known solutions and trying to
    return it.
    """

    def __init__(self, model: LmPredictor, include_lm_response: bool = False) -> None:
        super().__init__(model, include_lm_response)

    def solve(
        self,
        problem: CodeProblem,
        randomizing_seed: int | None = None,
    ) -> CodeSolution:
        return problem.known_solutions[0]


def snap_line_indent(string: str, indent: int) -> str:
    """Makes every line in the string snap to the nearest multiple of indent"""
    lines = string.split("\n")
    for i in range(len(lines)):
        stripped = lines[i].lstrip()
        current_indent = len(lines[i]) - len(stripped)
        new_indent = round(current_indent / indent) * indent
        lines[i] = " " * new_indent + stripped
    return "\n".join(lines)


class LmCodeSolverAutoRegressive(LmBasedSolver):
    """
    A code solver that works by prompting a left-to-right language model
    and parsing the response.
    """

    def __init__(
        self,
        model: LmPredictor,
        prompt_renderer: PromptRenderer = PromptRendererSingleEditGeneric(),
        response_parser: ResponseParser = ResponseParserSingleEdit(),
        prompting_test_case_strategy: PromptingTestCaseSelectionStrategy = None,
        include_lm_response: bool = False,
        snap_indent: int | None = None,
        cache_lm: bool = True,
    ) -> None:
        super().__init__(model, include_lm_response)
        self.prompt_renderer = prompt_renderer
        self.response_parser = response_parser
        self.snap_indent = snap_indent
        self.cache_lm = cache_lm
        # if isinstance(model, OpenAIPredictor) and include_lm_response:
        #    msg = (
        #        "OpenAI models have changed to not support echoing/logprobs and this"
        #        " now needs to be handled. Currently not implemented."
        #    )
        #    raise NotImplementedError(
        #        msg,
        #    )
        if prompting_test_case_strategy is not None:
            msg = "Prompting test case strategy not implemented refactored"
            raise NotImplementedError(
                msg,
            )

    def solve(
        self,
        problem: CodeProblem,
        randomizing_seed: int | None = None,
    ) -> LmCodeSolution:
        start = timer()
        render = self.prompt_renderer.render(
            problem=problem,
            lm=self.model,
            prompt_seed=None,
        )
        if isinstance(render, LmPromptRenderSingleEdit):
            if render.changed_element is None:
                msg = "Insert element cannot be None"
                raise ValueError(msg)
            if not isinstance(render.changed_element, MarkElement):
                msg = (
                    "Render error. Expected changed element to be a MarkElement. Got: "
                    + str(type(render.changed_element))
                )
                raise ValueError(msg)
            if not isinstance(render.changed_element.verb, StsPlaceTransforming):
                msg = (
                    "Render error. Expected changed element verb to be "
                    "a verb StsPlaceTransforming. Got: "
                    + str(type(render.changed_element.verb))
                )
                raise ValueError(msg)
        logprobs = render.prompt.logprobs
        if self.include_lm_response and (logprobs is None or logprobs == 0):
            logprobs = 1
        render = dataclasses.replace(
            render,
            prompt=dataclasses.replace(
                render.prompt,
                temperature=0,
                max_tokens=1000,
                cache=self.cache_lm,
                logprobs=logprobs,
                echo=False,  # Broken now with logprobs
            ),
        )

        resp = self.model.predict(render.prompt)
        solve_steps = self.response_parser.parse(render, resp, problem)
        if self.snap_indent:
            # TODO: move this into response parser?
            solve_steps = (
                dataclasses.replace(
                    solve_steps[0],
                    value=snap_line_indent(solve_steps[0].value, self.snap_indent),
                ),
            )
        end = timer()

        # Uncertainty stuff (TODO: refector)
        # if render.prompt.is_text_a_chat():
        #    mod_text = LmChatDialog(
        #        [
        #            *render.prompt.text,
        #            LmChatTurn(
        #                ChatGptRoles.assistant,
        #                solution.solve_steps[0].value,
        #            ),
        #        ],
        #    )
        # else:
        #    mod_text = render.code_prefix + solution.solve_steps[0].value
        # self._augment_solution_with_uncertainty_hooks(solution, resp, mod_text)
        lang_spec = PythonLangSpec()
        for step in solve_steps:
            if step.path is not None:
                lang_spec = problem.get_lang_spec_for_path(step.path)
                break

        return LmCodeSolution(
            solve_steps,
            problem,
            lang_spec=lang_spec,
            time=end - start,
            prompt=render.prompt,
            lm_prediction=resp if self.include_lm_response else None,
        )


class LmUncertaintyHook:
    def prepare(
        self,
        lm_prediction: LmPrediction,
        model: LmPredictor,
        solution: CodeSolution,
        modified_full_text: str,  # The prompt + completion (but with
        # possible changes to completion like truncation)
    ) -> DelayedUncertaintyEstimate:
        pass

    def get_name(self) -> str:
        return type(self).__name__

    def __eq__(self, other):
        return type(self).__name__ == type(other).__name__

    def __hash__(self):
        return hash(type(self).__name__)


class LmUncertaintyProbTotal(LmUncertaintyHook):
    def prepare(
        self,
        lm_prediction: LmPrediction,
        model: LmPredictor,
        solution: CodeSolution,
        modified_full_text: str,
    ) -> DelayedUncertaintyEstimate:
        start, end = estimate_best_overlap_span(
            lm_prediction.completion_text,
            larger_tokens=lm_prediction.completion_tokens,
            mod_string=solution.solve_steps[0].value,
            tokenizer_func=model.tokenize,
        )
        return PrecomputedUncertaintyEstimate[ProbabilityIsCorrect](
            ProbabilityIsCorrect(
                name_key="token_prob_total",
                value=np.exp(
                    np.sum(lm_prediction.completion_logprobs[start:end]),
                ),
            ),
        )

    def get_name(self) -> str:
        return "token_prob_total"


class LmUncertaintyProbAverage(LmUncertaintyHook):
    def prepare(
        self,
        lm_prediction: LmPrediction,
        model: LmPredictor,
        solution: CodeSolution,
        modified_full_text: str,
    ) -> DelayedUncertaintyEstimate:
        start, end = estimate_best_overlap_span(
            lm_prediction.completion_text,
            larger_tokens=lm_prediction.completion_tokens,
            mod_string=solution.solve_steps[0].value,
            tokenizer_func=model.tokenize,
        )
        out_estimate = ProbabilityIsCorrect(
            name_key="token_prob_avg",
            value=float(
                np.mean(np.exp(lm_prediction.completion_logprobs[start:end])),
            ),
        )
        return PrecomputedUncertaintyEstimate[ProbabilityIsCorrect](
            out_estimate,
        )

    def get_name(self) -> str:
        return "token_prob_avg"
